import copy
import hashlib
import uuid
from datetime import datetime

import json
from dateutil.parser import parse

import thundra.constants as constants
import thundra.utils as utils
from thundra.compat import PY37
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.encoder import to_json
from thundra.integrations.base_integration import BaseIntegration

OPERATION_TYPE_MAPPING_PATTERNS = utils.get_compiled_operation_type_patterns()


def dummy_func(*args):
    return None


def get_operation_type(class_name, operation_name):
    if class_name in constants.OperationTypeMappings["exclusions"] and \
            operation_name in constants.OperationTypeMappings["exclusions"][class_name]:
        return constants.OperationTypeMappings["exclusions"][class_name][operation_name]

    for sre in OPERATION_TYPE_MAPPING_PATTERNS:
        if sre.match(operation_name):
            return constants.OperationTypeMappings["patterns"][sre.pattern]
    return ''


class AWSDynamoDBIntegration(BaseIntegration):
    CLASS_TYPE = 'dynamodb'

    def __init__(self):
        self.OPERATION = {
            'PutItem': self.process_put_item_op,
            'UpdateItem': self.process_update_item_op,
            'GetItem': self.process_get_item_op,
            'DeleteItem': self.process_delete_item_op,
            'BatchWriteItem': self.process_batch_write_op,
        }

    def get_operation_name(self, wrapped, instance, args, kwargs):
        _, request_data = args
        return str(request_data['TableName']) if 'TableName' in request_data else constants.AWS_SERVICE_REQUEST

    def before_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        scope.span.domain_name = constants.DomainNames['DB']
        scope.span.class_name = constants.ClassNames['DYNAMODB']
        operation_name, request_data = args
        operation_type = get_operation_type(scope.span.class_name, operation_name)

        self.request_data = request_data.copy()
        self.endpoint = instance._endpoint.host.split('/')[-1]

        tags = {
            constants.SpanTags['OPERATION_TYPE']: operation_type,
            constants.DBTags['DB_INSTANCE']: self.endpoint,
            constants.DBTags['DB_TYPE']: constants.DBTypes['DYNAMODB'],
            constants.AwsDynamoTags['TABLE_NAME']: str(
                self.request_data['TableName']) if 'TableName' in self.request_data else None,
            constants.DBTags['DB_STATEMENT_TYPE']: operation_type,
            constants.AwsSDKTags['REQUEST_NAME']: operation_name,
        }
        scope.span.tags = tags

        # Check if Key and Item fields have any byte field and convert to string
        if 'Key' in self.request_data:
            self.escape_byte_fields(self.request_data['Key'])
        if 'Item' in self.request_data:
            self.escape_byte_fields(self.request_data['Item'])

        # DB statement tags should not be set on span if masked
        if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_AWS_DYNAMODB_STATEMENT_MASK):
            self.OPERATION.get(operation_name, dummy_func)(scope)

        scope.span.set_tag(constants.SpanTags['TOPOLOGY_VERTEX'], True)

        if ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_AWS_DYNAMODB_TRACEINJECTION_ENABLE):
            if operation_name == 'PutItem':
                self.inject_trace_link_on_put(scope.span, request_data, instance)

            if operation_name == 'UpdateItem':
                self.inject_trace_link_on_update(scope.span, request_data, instance)

            if operation_name == 'DeleteItem':
                self.inject_trace_link_on_delete(request_data)

    def after_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        super(AWSDynamoDBIntegration, self).after_call(scope, wrapped, instance, args, kwargs, response, exception)

        if scope.span.get_tag(constants.SpanTags['TRACE_LINKS']):
            return

        trace_links = self.get_trace_links(instance, args, response)
        if trace_links:
            scope.span.set_tag(constants.SpanTags['TRACE_LINKS'], trace_links)

    def get_trace_links(self, instance, args, response):
        operation_name, request_data = args
        trace_links = None

        try:
            params = copy.deepcopy(request_data)
            if PY37:
                id_handlers = instance.meta.events._emitter._unique_id_handlers
            else:
                id_handlers = instance.meta.events._unique_id_handlers

            if 'dynamodb-attr-value-input' in id_handlers:
                id_handlers['dynamodb-attr-value-input']['handler'](params=params,
                                                                    model=instance._service_model.operation_model(
                                                                        operation_name))

            region = instance.meta.region_name

            if operation_name == 'PutItem':
                trace_links = self.generate_trace_links(region, response, params['TableName'], 'SAVE',
                                                        params['Item'])

            elif operation_name == 'UpdateItem':
                trace_links = self.generate_trace_links(region, response, params['TableName'], 'SAVE',
                                                        params['Key'])

            elif operation_name == 'DeleteItem':
                if ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_AWS_DYNAMODB_TRACEINJECTION_ENABLE) and \
                        'Attributes' in response:
                    span_id = response['Attributes'].get(constants.THUNDRA_SPAN_ID_KEY)
                    if span_id and span_id.get('S'):
                        trace_links = ['DELETE:' + span_id.get('S')]

                if not trace_links:
                    trace_links = self.generate_trace_links(region, response, params['TableName'], 'DELETE',
                                                            params['Key'])

        except Exception:
            pass

        return trace_links

    def generate_trace_links(self, region, response, table_name, operation_type, attributes):
        try:
            date_str = response["ResponseMetadata"]["HTTPHeaders"]["date"]
            timestamp = (parse(date_str).replace(tzinfo=None) - datetime(1970, 1, 1)).total_seconds()
        except:
            timestamp = (datetime.now() - datetime(1970, 1, 1)).total_seconds() - 1

        attributes_hash = hashlib.md5(self.attributes_to_str(attributes).encode()).hexdigest()

        trace_links = [
            region + ':' + table_name + ':' + str(int(timestamp + i)) + ':' + operation_type + ':' + attributes_hash for
            i in range(3)]
        return trace_links

    @staticmethod
    def attributes_to_str(attributes):
        sorted_keys = sorted(attributes.keys())
        attributes_sorted = []
        for attr in sorted_keys:
            try:
                key = list(attributes[attr].keys())[0]
                attributes_sorted.append(attr + '=' + '{' + key + ': ' + str(attributes[attr][key]) + '}')
            except Exception as e:
                pass
        return ', '.join(attributes_sorted)

    def inject_trace_link_on_put(self, span, request_data, instance):

        thundra_span = {'S': span.span_id}
        try:
            if PY37:
                id_handlers = instance.meta.events._emitter._unique_id_handlers
            else:
                id_handlers = instance.meta.events._unique_id_handlers

            if 'dynamodb-attr-value-input' in id_handlers:
                thundra_span = span.span_id
        except Exception as e:
            pass

        try:
            if 'Item' in request_data:
                request_data['Item'][constants.THUNDRA_SPAN_ID_KEY] = thundra_span
            else:
                request_data['Item'] = {constants.THUNDRA_SPAN_ID_KEY: thundra_span}

            span.set_tag(constants.SpanTags['TRACE_LINKS'], ["SAVE:" + span.span_id])
        except:
            pass

    def inject_trace_link_on_update(self, span, request_data, instance):

        thundra_span = {'S': span.span_id}
        try:
            if PY37:
                id_handlers = instance.meta.events._emitter._unique_id_handlers
            else:
                id_handlers = instance.meta.events._unique_id_handlers
            if 'dynamodb-attr-value-input' in id_handlers:
                thundra_span = span.span_id
        except Exception as e:
            pass

        thundra_attr = {'Action': 'PUT', 'Value': thundra_span}

        try:
            if 'AttributeUpdates' in request_data:
                request_data['AttributeUpdates'][constants.THUNDRA_SPAN_ID_KEY] = thundra_attr
            else:
                request_data['AttributeUpdates'] = {constants.THUNDRA_SPAN_ID_KEY: thundra_attr}

            span.set_tag(constants.SpanTags['TRACE_LINKS'], ["SAVE:" + span.span_id])
        except:
            pass

    def inject_trace_link_on_delete(self, request_data):
        try:
            request_data['ReturnValues'] = "ALL_OLD"
        except:
            pass

    def process_get_item_op(self, scope):
        if 'Key' in self.request_data:
            scope.span.set_tag(constants.DBTags['DB_STATEMENT'], self.request_data['Key'])

    def process_put_item_op(self, scope):
        if 'Item' in self.request_data:
            scope.span.set_tag(constants.DBTags['DB_STATEMENT'], self.request_data['Item'])

    def process_delete_item_op(self, scope):
        if 'Key' in self.request_data:
            scope.span.set_tag(constants.DBTags['DB_STATEMENT'], self.request_data['Key'])

    def process_update_item_op(self, scope):
        if 'Key' in self.request_data:
            scope.span.set_tag(constants.DBTags['DB_STATEMENT'], self.request_data['Key'])

    def escape_byte_fields(self, req_dict):
        for k, v in req_dict.items():
            if type(v) == dict:
                for f, fv in v.items():
                    if type(fv) == bytes:
                        v[f] = fv.decode()
                    elif type(fv) == list and len(fv) != 0 and type(fv[0]) == bytes:
                        v[f] = [e.decode() for e in fv]
            elif type(v) == bytes:
                req_dict[k] = v.decode()
            elif type(v) == list and len(v) != 0 and type(v[0]) == bytes:
                req_dict[k] = [e.decode() for e in v]

    def process_batch_write_op(self, scope):
        table_name = list(self.request_data['RequestItems'].keys())[0]
        items = []
        for item in self.request_data['RequestItems'][table_name]:
            items.append(item)

        scope.span.set_tag(constants.DBTags['DB_STATEMENT'], items)


class AWSSQSIntegration(BaseIntegration):
    CLASS_TYPE = 'sqs'

    def __init__(self):
        pass

    def get_operation_name(self, wrapped, instance, args, kwargs):
        _, request_data = args
        queue_name = str(self.getQueueName(request_data))
        if queue_name != '':
            return queue_name
        return constants.AWS_SERVICE_REQUEST

    def getQueueName(self, data):
        if 'QueueUrl' in data:
            return data['QueueUrl'].split('/')[-1]
        elif 'QueueName' in data:
            return data['QueueName']
        return ''

    def before_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        operation_name, request_data = args
        self.request_data = request_data
        self.queueName = str(self.getQueueName(self.request_data))

        scope.span.domain_name = constants.DomainNames['MESSAGING']
        scope.span.class_name = constants.ClassNames['SQS']

        tags = {
            constants.AwsSQSTags['QUEUE_NAME']: self.queueName,
            constants.SpanTags['OPERATION_TYPE']: get_operation_type(scope.span.class_name, operation_name),
            constants.AwsSDKTags['REQUEST_NAME']: operation_name,
        }

        scope.span.tags = tags
        scope.span.set_tag(constants.SpanTags['TOPOLOGY_VERTEX'], True)

        if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_AWS_SQS_MESSAGE_MASK):
            if operation_name == "SendMessage":
                message = request_data.get("MessageBody", "")
                scope.span.set_tag(constants.AwsSQSTags['MESSAGE'], message)

            elif operation_name == "SendMessageBatch":
                entries = request_data.get('Entries', None)
                messages = []
                if entries:
                    for entry in entries:
                        messages.append(entry.get("MessageBody", ""))
                scope.span.set_tag(constants.AwsSQSTags['MESSAGES'], messages)

    def after_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        super(AWSSQSIntegration, self).after_call(scope, wrapped, instance, args, kwargs, response, exception)
        operation_name = args[0]

        trace_links = self.get_trace_links(operation_name, response)
        if trace_links:
            scope.span.set_tag(constants.SpanTags['TRACE_LINKS'], trace_links)

    @staticmethod
    def get_trace_links(operation_name, response):
        if operation_name == 'SendMessage':
            message_id = response['MessageId']
            return [message_id]

        elif operation_name == 'SendMessageBatch':
            if len(response['Successful']) > 0:
                links = []
                entries = response['Successful']
                for entry in entries:
                    links.append(entry['MessageId'])
                return links
        return None


class AWSSNSIntegration(BaseIntegration):
    CLASS_TYPE = 'sns'

    def __init__(self):
        pass

    def get_operation_name(self, wrapped, instance, args, kwargs):
        operation_name, request_data = args
        if operation_name == 'CreateTopic':
            self.topicName = request_data.get('Name', 'N/A')
        else:
            arn = request_data.get(
                'TopicArn',
                request_data.get('TargetArn', 'N/A')
            )
            if arn != 'N/A':
                self.topicName = arn.split(':')[-1]
            else:
                self.topicName = ''

        return self.topicName

    def before_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        operation_name, request_data = args
        self.request_data = request_data
        self.response = response
        self.message = request_data.get('Message', '')

        scope.span.domain_name = constants.DomainNames['MESSAGING']
        scope.span.class_name = constants.ClassNames['SNS']

        tags = {
            constants.AwsSDKTags['REQUEST_NAME']: operation_name,
            constants.SpanTags['OPERATION_TYPE']: get_operation_type(scope.span.class_name, operation_name),
            constants.AwsSNSTags['TOPIC_NAME']: self.topicName
        }

        scope.span.tags = tags
        scope.span.set_tag(constants.SpanTags['TOPOLOGY_VERTEX'], True)

        if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_AWS_SNS_MESSAGE_MASK):
            scope.span.set_tag(constants.AwsSNSTags['MESSAGE'], self.message)

    def after_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        super(AWSSNSIntegration, self).after_call(scope, wrapped, instance, args, kwargs, response, exception)

        trace_links = self.get_trace_links(response)
        if trace_links:
            scope.span.set_tag(constants.SpanTags['TRACE_LINKS'], trace_links)

    @staticmethod
    def get_trace_links(response):
        try:
            message_id = response['MessageId']
            return [message_id]
        except Exception:
            return None


class AWSKinesisIntegration(BaseIntegration):
    CLASS_TYPE = 'kinesis'

    def __init__(self):
        pass

    def get_operation_name(self, wrapped, instance, args, kwargs):
        _, request_data = args
        return request_data.get('StreamName', constants.AWS_SERVICE_REQUEST)

    def generate_trace_link(self, region, shard_id, sequence_number):
        return region + ':' + self.streamName + ':' + shard_id + ':' + sequence_number

    def before_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        operation_name, request_data = args
        self.streamName = request_data.get('StreamName', '')

        scope.span.domain_name = constants.DomainNames['STREAM']
        scope.span.class_name = constants.ClassNames['KINESIS']

        tags = {
            constants.AwsSDKTags['REQUEST_NAME']: operation_name,
            constants.SpanTags['OPERATION_TYPE']: get_operation_type(scope.span.class_name, operation_name),
            constants.AwsKinesisTags['STREAM_NAME']: self.streamName
        }

        scope.span.tags = tags
        scope.span.set_tag(constants.SpanTags['TOPOLOGY_VERTEX'], True)

    def after_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        super(AWSKinesisIntegration, self).after_call(scope, wrapped, instance, args, kwargs, response, exception)

        trace_links = self.get_trace_links(instance.meta.region_name, response)
        scope.span.set_tag(constants.SpanTags['TRACE_LINKS'], trace_links)

    def get_trace_links(self, region, response):
        trace_links = []
        try:
            if "Records" in response:
                records = response["Records"]
                for record in records:
                    trace_links.append(
                        self.generate_trace_link(region, record["ShardId"], record["SequenceNumber"]))
            else:
                trace_links.append(
                    self.generate_trace_link(region, response["ShardId"], response["SequenceNumber"]))
        except:
            return None

        return trace_links


class AWSFirehoseIntegration(BaseIntegration):
    CLASS_TYPE = 'firehose'

    def __init__(self):
        pass

    def get_operation_name(self, wrapped, instance, args, kwargs):
        _, request_data = args
        return request_data.get('DeliveryStreamName', constants.AWS_SERVICE_REQUEST)

    def generate_trace_links(self, region, response, data):
        try:
            date_str = response["ResponseMetadata"]["HTTPHeaders"]["date"]
            timestamp = (parse(date_str).replace(tzinfo=None) - datetime(1970, 1, 1)).total_seconds()

        except:
            timestamp = (datetime.now() - datetime(1970, 1, 1)).total_seconds() - 1

        if isinstance(data, str):
            data = data.encode()
        data_md5 = hashlib.md5(data).hexdigest()

        trace_links = [region + ':' + self.deliveryStreamName + ':' + str(int(timestamp + i)) + ':' + data_md5 for i in
                       range(3)]
        return trace_links

    def before_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        operation_name, request_data = args
        self.deliveryStreamName = request_data.get('DeliveryStreamName', '')

        scope.span.domain_name = constants.DomainNames['STREAM']
        scope.span.class_name = constants.ClassNames['FIREHOSE']

        tags = {
            constants.AwsSDKTags['REQUEST_NAME']: operation_name,
            constants.SpanTags['OPERATION_TYPE']: get_operation_type(scope.span.class_name, operation_name),
            constants.AwsFirehoseTags['STREAM_NAME']: self.deliveryStreamName
        }

        scope.span.tags = tags
        scope.span.set_tag(constants.SpanTags['TOPOLOGY_VERTEX'], True)

    def after_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        super(AWSFirehoseIntegration, self).after_call(scope, wrapped, instance, args, kwargs, response, exception)

        trace_links = self.get_trace_links(args, instance.meta.region_name, response)
        if trace_links:
            scope.span.set_tag(constants.SpanTags['TRACE_LINKS'], trace_links)

    def get_trace_links(self, args, region, response):
        trace_links = []
        try:
            operation_name, request_data = args

            if operation_name == "PutRecord":
                data = request_data["Record"]["Data"]
                trace_links = self.generate_trace_links(region, response, data)

            elif operation_name == "PutRecordBatch":
                for record in request_data["Records"]:
                    links = self.generate_trace_links(region, response, record["Data"])
                    trace_links.extend(links)

        except Exception as e:
            pass

        return trace_links if len(trace_links) > 0 else None


class AWSS3Integration(BaseIntegration):
    CLASS_TYPE = 's3'

    def __init__(self):
        pass

    def get_operation_name(self, wrapped, instance, args, kwargs):
        _, request_data = args
        return request_data['Bucket'] if 'Bucket' in request_data else constants.AWS_SERVICE_REQUEST

    def before_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        operation_name, request_data = args
        self.bucket = request_data['Bucket'] if 'Bucket' in request_data else ''

        scope.span.domain_name = constants.DomainNames['STORAGE']
        scope.span.class_name = constants.ClassNames['S3']

        self.objectName = request_data.get('Key', '')

        tags = {
            constants.AwsSDKTags['REQUEST_NAME']: operation_name,
            constants.SpanTags['OPERATION_TYPE']: get_operation_type(scope.span.class_name, operation_name),
            constants.AwsS3Tags['BUCKET_NAME']: self.bucket,
            constants.AwsS3Tags['OBJECT_NAME']: self.objectName
        }

        scope.span.tags = tags
        scope.span.set_tag(constants.SpanTags['TOPOLOGY_VERTEX'], True)

    def after_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        super(AWSS3Integration, self).after_call(scope, wrapped, instance, args, kwargs, response, exception)

        trace_links = self.get_trace_links(response)
        if trace_links:
            scope.span.set_tag(constants.SpanTags['TRACE_LINKS'], trace_links)

    def get_trace_links(self, response):
        try:
            request_id = response["ResponseMetadata"]["HTTPHeaders"]['x-amz-request-id']
            return [request_id]
        except Exception:
            return None


class AWSLambdaIntegration(BaseIntegration):
    CLASS_TYPE = 'lambda'

    def __init__(self):
        pass

    def get_operation_name(self, wrapped, instance, args, kwargs):
        _, request_data = args
        operation_name = constants.AWS_SERVICE_REQUEST
        if request_data.get('FunctionName'):
            operation_name = self.normalize_function_name(request_data.get('FunctionName')).get('name', operation_name)
        return operation_name

    def before_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        operation_name, request_data = args
        lambda_function = request_data.get('FunctionName', '')
        scope.span.domain_name = constants.DomainNames['API']
        scope.span.class_name = constants.ClassNames['LAMBDA']

        lambda_function = self.normalize_function_name(lambda_function)

        tags = {
            constants.AwsSDKTags['REQUEST_NAME']: operation_name,
            constants.SpanTags['OPERATION_TYPE']: get_operation_type(scope.span.class_name, operation_name),
            constants.AwsLambdaTags['FUNCTION_NAME']: lambda_function.get('name'),
            constants.AwsLambdaTags['FUNCTION_QUALIFIER']: lambda_function.get('qualifier')
        }

        if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_AWS_LAMBDA_PAYLOAD_MASK) and \
                'Payload' in request_data:
            tags[constants.AwsLambdaTags['INVOCATION_PAYLOAD']] = str(request_data['Payload'],
                                                                      encoding='utf-8') if type(
                request_data['Payload']) is not str else request_data['Payload']

        if 'Qualifier' in request_data:
            tags[constants.AwsLambdaTags['FUNCTION_QUALIFIER']] = request_data['Qualifier']

        if 'InvocationType' in request_data:
            tags[constants.AwsLambdaTags['INVOCATION_TYPE']] = request_data['InvocationType']

        scope.span.tags = tags
        scope.span.set_tag(constants.SpanTags['TOPOLOGY_VERTEX'], True)

    def after_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        super(AWSLambdaIntegration, self).after_call(scope, wrapped, instance, args, kwargs, response, exception)

        trace_links = self.get_trace_links(response)
        if trace_links:
            scope.span.set_tag(constants.SpanTags['TRACE_LINKS'], trace_links)

    @staticmethod
    def get_trace_links(response):
        try:
            request_id = response["ResponseMetadata"]["HTTPHeaders"]['x-amzn-requestid']
            return [request_id]
        except Exception:
            return None

    @staticmethod
    def normalize_function_name(function_name):
        parts = function_name.split(':')
        if len(parts) < 2:  # funcName
            return {'name': function_name}
        if len(parts) == 2:  # funcName:qualifier
            return {'name': parts[0], 'qualifier': parts[1]}
        if len(parts) == 3:  # accountId:function:funcName
            return {'name': parts[2]}
        if len(parts) == 4:  # accountId:function:funcName:qualifier
            return {'name': parts[2], 'qualifier': parts[2]}
        if len(parts) == 7:  # arn:aws:lambda:region:accountId:function:funcName
            return {'name': parts[6]}
        if len(parts) == 8:  # arn:aws:lambda:region:accountId:function:funcName:qualifier
            return {'name': parts[6], 'qualifier': parts[7]}
        return {}


class AWSServiceIntegration(BaseIntegration):
    CLASS_TYPE = 'default'

    def __init__(self):
        pass

    def get_operation_name(self, wrapped, instance, args, kwargs):
        return constants.AWS_SERVICE_REQUEST

    def before_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        scope.span.domain_name = constants.DomainNames['AWS']
        scope.span.class_name = constants.ClassNames['AWSSERVICE']

        service_name = instance.__class__.__name__.lower()

        if len(args) > 0:
            scope.span.set_tag(constants.AwsSDKTags['REQUEST_NAME'], args[0])
            scope.span.set_tag(constants.SpanTags['OPERATION_TYPE'], get_operation_type(scope.span.class_name, args[0]))

        scope.span.set_tag(constants.AwsSDKTags['SERVICE_NAME'], service_name)


class AWSStepFunctionIntegration(BaseIntegration):
    CLASS_TYPE = 'sfn'

    def __init__(self):
        pass

    def get_operation_name(self, wrapped, instance, args, kwargs):
        _, request_data = args
        state_machine_arn = request_data.get('stateMachineArn')
        if state_machine_arn:
            return state_machine_arn.split(':')[-1]
        return constants.AWS_SERVICE_REQUEST

    def before_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        scope.span.domain_name = constants.DomainNames['AWS']
        scope.span.class_name = constants.ClassNames['STEPFUNCTIONS']

        _, request_data = args
        state_machine_arn = request_data.get('stateMachineArn', '')
        execution_name = request_data.get('name', '')

        service_name = instance.__class__.__name__.lower()

        try:
            orig_input = request_data.get('input', None)
            if orig_input != None and ConfigProvider.get(config_names.THUNDRA_LAMBDA_AWS_STEPFUNCTIONS):
                parsed_input = json.loads(orig_input)
                trace_link = str(uuid.uuid4())
                parsed_input['_thundra'] = {
                    "trace_link": trace_link,
                    "step": 0
                }
                scope.span.set_tag(constants.AwsStepFunctionsTags['EXECUTION_INPUT'], orig_input)
                request_data['input'] = to_json(parsed_input)
                scope.span.set_tag(constants.SpanTags['TRACE_LINKS'], [trace_link])
                scope.span.resource_trace_links = [trace_link]
        except:
            pass

        if len(args) > 0:
            scope.span.set_tag(constants.AwsSDKTags['REQUEST_NAME'], args[0])
            scope.span.set_tag(constants.SpanTags['OPERATION_TYPE'], get_operation_type(scope.span.class_name, args[0]))

        scope.span.set_tag(constants.AwsSDKTags['SERVICE_NAME'], service_name)
        scope.span.set_tag(constants.AwsStepFunctionsTags['STATE_MACHINE_ARN'], state_machine_arn)
        scope.span.set_tag(constants.AwsStepFunctionsTags['EXECUTION_NAME'], execution_name)

        scope.span.set_tag(constants.SpanTags['TOPOLOGY_VERTEX'], True)

    def after_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        super(AWSStepFunctionIntegration, self).after_call(scope, wrapped, instance, args, kwargs, response, exception)
        scope.span.set_tag(constants.AwsStepFunctionsTags['EXECUTION_ARN'], response.get('executionArn', ''))
        try:
            if response.get('startDate'):
                start_date_str = response.get('startDate').isoformat()
                scope.span.set_tag(constants.AwsStepFunctionsTags['EXECUTION_START_DATE'], start_date_str)
        except:
            pass


class AWSAthenaIntegration(BaseIntegration):
    CLASS_TYPE = 'athena'

    def __init__(self):
        pass

    def get_operation_name(self, wrapped, instance, args, kwargs):
        return self.get_database_name(args) or constants.AWS_SERVICE_REQUEST

    def get_database_name(self, args):
        try:
            database = args[1]['QueryExecutionContext']['Database']
        except:
            database = None
        return database

    def get_output_location(self, args):
        try:
            output_location = args[1]['ResultConfiguration']['OutputLocation']
        except:
            output_location = None
        return output_location

    def get_query(self, args):
        try:
            query = args[1]['QueryString']
        except:
            query = None
        return query

    def get_query_execution_ids(self, args):
        query_execution_ids = None
        if len(args) > 1:
            params = args[1]

            if params and "QueryExecutionId" in params:
                query_execution_ids = [params.get("QueryExecutionId")]
            elif params and "QueryExecutionIds" in params:
                query_execution_ids = params.get("QueryExecutionIds")
        return query_execution_ids

    def get_named_query_ids(self, args):
        named_query_ids = None
        if len(args) > 1:
            params = args[1]

            if params and "NamedQueryId" in params:
                named_query_ids = [params.get("NamedQueryId")]
            elif params and "NamedQueryIds" in params:
                named_query_ids = params.get("NamedQueryIds")
        return named_query_ids

    def before_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        operation_name = args[0]
        scope.span.domain_name = constants.DomainNames['DB']
        scope.span.class_name = constants.ClassNames['ATHENA']

        tags = {
            constants.SpanTags['TOPOLOGY_VERTEX']: True,
            constants.AwsSDKTags['REQUEST_NAME']: operation_name,
            constants.SpanTags['OPERATION_TYPE']: get_operation_type(scope.span.class_name, operation_name),
        }

        scope.span.tags = tags

        database = self.get_database_name(args)
        output_location = self.get_output_location(args)
        query_execution_ids = self.get_query_execution_ids(args)
        named_query_ids = self.get_named_query_ids(args)

        if database:
            scope.span.set_tag(constants.SpanTags['DB_INSTANCE'], database)

        if output_location:
            scope.span.set_tag(constants.AthenaTags['S3_OUTPUT_LOCATION'], output_location)

        if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_AWS_ATHENA_STATEMENT_MASK):
            scope.span.set_tag(constants.DBTags['DB_STATEMENT'], self.get_query(args))

        if query_execution_ids:
            scope.span.set_tag(constants.AthenaTags['REQUEST_QUERY_EXECUTION_IDS'], query_execution_ids)

        if named_query_ids:
            scope.span.set_tag(constants.AthenaTags['REQUEST_NAMED_QUERY_IDS'], named_query_ids)

    def after_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        if response:
            if "QueryExecutionId" in response:
                scope.span.set_tag(constants.AthenaTags['RESPONSE_QUERY_EXECUTION_IDS'],
                                   [response.get("QueryExecutionId")])
            elif "QueryExecutionIds" in response:
                scope.span.set_tag(constants.AthenaTags['RESPONSE_QUERY_EXECUTION_IDS'],
                                   response.get("QueryExecutionIds"))
            elif "NamedQueryId" in response:
                scope.span.set_tag(constants.AthenaTags['RESPONSE_NAMED_QUERY_IDS'], [response.get("NamedQueryId")])
            elif "NamedQueryIds" in response:
                scope.span.set_tag(constants.AthenaTags['RESPONSE_NAMED_QUERY_IDS'], response.get("NamedQueryIds"))


class AWSEventBridgeIntegration(BaseIntegration):
    CLASS_TYPE = 'eventbridge'

    def __init__(self):
        pass


    def get_operation_name(self, wrapped, instance, args, kwargs):
        _, request_data = args
        operation_name = constants.AwsEventBridgeTags['SERVICE_REQUEST']
        entries = request_data.get('Entries', [])
        eventbus_map = set()
        for entry in entries:
            op_name = self.extract_operation_name(entry)
            if op_name != None:
                eventbus_map.add(op_name)
        if len(eventbus_map) == 1:
            operation_name = eventbus_map.pop()
        return operation_name


    def extract_operation_name(self, entry):
        return entry.get("EventBusName", None) or entry.get("DetailType", None) or entry.get("Source", None)


    def before_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        operation_name, request_data = args
        scope.span.domain_name = constants.DomainNames['MESSAGING']
        scope.span.class_name = constants.ClassNames['EVENTBRIDGE']

        tags = {
            constants.AwsSDKTags['REQUEST_NAME']: operation_name,
            constants.SpanTags['OPERATION_TYPE']: get_operation_type(scope.span.class_name, operation_name)
        }

        entries = []
        for entry in request_data.get('Entries', []):
            new_entry = {
                'Detail': None if ConfigProvider.get(
                    config_names.THUNDRA_TRACE_INTEGRATIONS_EVENTBRIDGE_DETAIL_MASK) else entry.get('Detail'),
                'DetailType': entry.get('DetailType'),
                'EventBusName': entry.get('EventBusName'),
                'Resources': entry.get('Resources'),
                'Source': entry.get('Source')
            }
            if isinstance(entry.get('Time'), datetime):
                new_entry['Time'] = datetime.timestamp(entry.get('Time'))
            elif isinstance(entry.get('Time'), int):
                new_entry['Time'] = entry.get('Time')
            entries.append(new_entry)

        if entries:
            tags[constants.AwsEventBridgeTags['ENTRIES']] = entries
            tags[constants.SpanTags['RESOURCE_NAMES']] = list(map(lambda x: x.get('DetailType'), entries))

        scope.span.tags = tags
        scope.span.set_tag(constants.SpanTags['TOPOLOGY_VERTEX'], True)

    def after_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        super(AWSEventBridgeIntegration, self).after_call(scope, wrapped, instance, args, kwargs, response, exception)

        trace_links = self.get_trace_links(response)
        if trace_links:
            scope.span.set_tag(constants.SpanTags['TRACE_LINKS'], trace_links)

    def get_trace_links(self, response):
        try:
            entries = response['Entries']
            event_ids = []
            for entry in entries:
                if entry.get('EventId'):
                    event_ids.append(entry.get('EventId'))
            return event_ids
        except Exception:
            return None


class AWSSESIntegration(BaseIntegration):
    CLASS_TYPE = 'ses'

    def __init__(self):
        pass

    def get_operation_name(self, wrapped, instance, args, kwargs):
        operation_name, request_data = args
        return operation_name if operation_name else constants.AWS_SERVICE_REQUEST

    def before_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        operation_name, request_data = args
        scope.span.domain_name = constants.DomainNames['MESSAGING']
        scope.span.class_name = constants.ClassNames['SES']

        mask_mail = ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_AWS_SES_MAIL_MASK)
        mask_destination = ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_AWS_SES_MAIL_DESTINATION_MASK)

        source = request_data.get('Source', '')
        destination = request_data.get('Destinations', request_data.get('Destination', []))
        subject = request_data.get('Message', {}).get('Subject')
        body = request_data.get('Message', {}).get('Body')
        template_name = request_data.get('Template')
        template_arn = request_data.get('TemplateArn')
        template_data = request_data.get('TemplateData')

        scope.span.tags = {
            constants.AwsSDKTags['REQUEST_NAME']: operation_name,
            constants.SpanTags['OPERATION_TYPE']: get_operation_type(scope.span.class_name, operation_name),
            constants.AwsSESTags['SOURCE']: source,
            constants.AwsSESTags['DESTINATION']: None if mask_destination else destination,
            constants.AwsSESTags['SUBJECT']: None if mask_mail else subject,
            constants.AwsSESTags['BODY']: None if mask_mail else body,
            constants.AwsSESTags['TEMPLATE_NAME']: template_name,
            constants.AwsSESTags['TEMPLATE_ARN']: template_arn,
            constants.AwsSESTags['TEMPLATE_DATA']: None if mask_mail else template_data
        }
        scope.span.set_tag(constants.SpanTags['TOPOLOGY_VERTEX'], True)

    def after_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        super(AWSSESIntegration, self).after_call(scope, wrapped, instance, args, kwargs, response, exception)



from thundra.context.context_provider import ContextProvider
from thundra.context.execution_context import ExecutionContext

_execution_context = ExecutionContext()


class GlobalExecutionContextProvider(ContextProvider):
    def get(self):
        return _execution_context

    def set(self, execution_context):
        global _execution_context
        _execution_context = execution_context

    def clear(self):
        global _execution_context
        _execution_context = ExecutionContext()



from thundra.samplers.base_sampler import BaseSampler


class ErrorAwareSampler(BaseSampler):

    def __init__(self):
        pass

    def is_sampled(self, span=None):
        if not span:
            return False

        return span.get_tag("error")



from .base_sampler import BaseSampler
from .duration_aware_sampler import DurationAwareSampler
from .error_aware_sampler import ErrorAwareSampler

from .time_aware_sampler import TimeAwareSampler
from .count_aware_sampler import CountAwareSampler
from .composite_sampler import CompositeSampler



from thundra.context.execution_context import ExecutionContext
from foresight.test_runner_tags import TestRunnerTags

class TestCaseExecutionContext(ExecutionContext):

    def __init__(self, **opts):
        super(TestCaseExecutionContext, self).__init__(**opts)
        self.node_id = opts.get("node_id", "")
        self.name = opts.get("name", "")
        self.method = opts.get("method", "")
        self.test_class = opts.get("test_class", '')
        self.test_suite_name = opts.get("test_suite_name", "")
        self.parent_transaction_id = opts.get("parent_transaction_id", "")
        self.status = opts.get("status", "")

    def set_status(self, status):
        self.status = status

    def get_operation_name(self):
        return "RunTest"

    def get_additional_start_tags(self):
        from foresight.test_runner_support import TestRunnerSupport
        test_run_scope = TestRunnerSupport.test_run_scope
        return {
            TestRunnerTags.TEST_RUN_ID: test_run_scope.id,
            TestRunnerTags.TEST_RUN_TASK_ID: test_run_scope.task_id,
            TestRunnerTags.TEST_NAME: self.name,
            TestRunnerTags.TEST_SUITE: self.test_suite_name,
            TestRunnerTags.TEST_METHOD: self.method,
            TestRunnerTags.TEST_CLASS: self.test_class,
            TestRunnerTags.TEST_SUITE_TRANSACTION_ID: self.parent_transaction_id
        }

    def get_additional_finish_tags(self):
        return {
            TestRunnerTags.TEST_STATUS: self.status
        }


from thundra.plugins.config.base_plugin_config import BasePluginConfig


class LogConfig(BasePluginConfig):
    def __init__(self, enabled=False, sampler=None):
        super(LogConfig, self).__init__(enabled=enabled)
        self.sampler = sampler



from __future__ import unicode_literals

import base64
import hashlib

import mock
import pytest
import simplejson as json

from thundra.compat import str
from thundra.context.execution_context_manager import ExecutionContextManager

try:
    from contextlib import ExitStack
except ImportError:
    from contextlib2 import ExitStack

from thundra import constants
from thundra.wrappers.aws_lambda import lambda_event_utils
from thundra.opentracing.tracer import ThundraTracer
from thundra.plugins import invocation

try:
    from BytesIO import BytesIO
except ImportError:
    from io import BytesIO
from gzip import GzipFile


@pytest.fixture
def tracer_and_invocation_support():
    with ExitStack() as stack:
        # Just a fancy way of using contexts to avoid an ugly multi-line with statement
        stack.enter_context(mock.patch('thundra.opentracing.recorder.ThundraRecorder.clear'))
        tracer = ThundraTracer.get_instance()
        invocation_support = invocation.invocation_support
        invocation_trace_support = invocation.invocation_trace_support
        yield tracer, invocation_support, invocation_trace_support


def test_dynamodb_trigger(tracer_and_invocation_support, handler, mock_dynamodb_event, mock_context):
    _, handler = handler
    tracer, invocation_support, invocation_trace_support = tracer_and_invocation_support
    handler(mock_dynamodb_event, mock_context)

    execution_context = ExecutionContextManager.get()
    span = execution_context.recorder.get_spans()[0]
    assert lambda_event_utils.get_lambda_event_type(mock_dynamodb_event,
                                                    mock_context) == lambda_event_utils.LambdaEventType.DynamoDB

    assert span.get_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == constants.DomainNames['DB']
    assert span.get_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == constants.ClassNames['DYNAMODB']
    assert span.get_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['ExampleTableWithStream']

    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == constants.DomainNames['DB']
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == constants.ClassNames[
        'DYNAMODB']
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['ExampleTableWithStream']

    md5_key = hashlib.md5("Id={N: 101}".encode()).hexdigest()
    md5_image_1 = hashlib.md5("Id={N: 101}, Message={S: New item!}".encode()).hexdigest()
    md5_image_2 = hashlib.md5("Id={N: 101}, Message={S: This item has changed}".encode()).hexdigest()
    region = 'eu-west-2'
    table_name = 'ExampleTableWithStream'
    timestamp = 1480642019

    links = [
        region + ':' + table_name + ':' + str(timestamp) + ':' + 'SAVE' + ':' + md5_key,
        region + ':' + table_name + ':' + str(timestamp + 1) + ':' + 'SAVE' + ':' + md5_key,
        region + ':' + table_name + ':' + str(timestamp + 2) + ':' + 'SAVE' + ':' + md5_key,
        region + ':' + table_name + ':' + str(timestamp) + ':' + 'SAVE' + ':' + md5_image_1,
        region + ':' + table_name + ':' + str(timestamp + 1) + ':' + 'SAVE' + ':' + md5_image_1,
        region + ':' + table_name + ':' + str(timestamp + 2) + ':' + 'SAVE' + ':' + md5_image_1,
        region + ':' + table_name + ':' + str(timestamp) + ':' + 'SAVE' + ':' + md5_image_2,
        region + ':' + table_name + ':' + str(timestamp + 1) + ':' + 'SAVE' + ':' + md5_image_2,
        region + ':' + table_name + ':' + str(timestamp + 2) + ':' + 'SAVE' + ':' + md5_image_2
    ]

    assert sorted(invocation_trace_support.get_incoming_trace_links().get('incomingTraceLinks')) == sorted(links)


def test_dynamodb_trigger_delete_event(tracer_and_invocation_support, handler, mock_dynamodb_delete_event,
                                       mock_context):
    _, handler = handler
    tracer, invocation_support, invocation_trace_support = tracer_and_invocation_support
    handler(mock_dynamodb_delete_event, mock_context)
    execution_context = ExecutionContextManager.get()
    span = execution_context.recorder.get_spans()[0]
    assert lambda_event_utils.get_lambda_event_type(mock_dynamodb_delete_event,
                                                    mock_context) == lambda_event_utils.LambdaEventType.DynamoDB

    assert span.get_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == constants.DomainNames['DB']
    assert span.get_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == constants.ClassNames['DYNAMODB']
    assert span.get_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['ExampleTableWithStream']

    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == constants.DomainNames['DB']
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == constants.ClassNames[
        'DYNAMODB']
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['ExampleTableWithStream']

    md5_key = hashlib.md5("Id={N: 101}".encode()).hexdigest()
    region = 'eu-west-2'
    table_name = 'ExampleTableWithStream'
    timestamp = 1480642019

    links = [
        region + ':' + table_name + ':' + str(timestamp) + ':' + 'DELETE' + ':' + md5_key,
        region + ':' + table_name + ':' + str(timestamp + 1) + ':' + 'DELETE' + ':' + md5_key,
        region + ':' + table_name + ':' + str(timestamp + 2) + ':' + 'DELETE' + ':' + md5_key
    ]
    assert sorted(invocation_trace_support.get_incoming_trace_links().get('incomingTraceLinks')) == sorted(links)


def test_dynamodb_trigger_trace_injected(tracer_and_invocation_support, handler, mock_dynamodb_event_trace_injected,
                                         mock_context):
    _, handler = handler
    tracer, invocation_support, invocation_trace_support = tracer_and_invocation_support
    handler(mock_dynamodb_event_trace_injected, mock_context)

    execution_context = ExecutionContextManager.get()
    span = execution_context.recorder.get_spans()[0]

    assert lambda_event_utils.get_lambda_event_type(mock_dynamodb_event_trace_injected,
                                                    mock_context) == lambda_event_utils.LambdaEventType.DynamoDB

    assert span.get_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == constants.DomainNames['DB']
    assert span.get_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == constants.ClassNames['DYNAMODB']
    assert span.get_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['ExampleTableWithStream']

    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == constants.DomainNames['DB']
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == constants.ClassNames[
        'DYNAMODB']
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['ExampleTableWithStream']

    assert sorted(invocation_trace_support.get_incoming_trace_links().get('incomingTraceLinks')) == sorted(
        ['SAVE:test_id1', 'SAVE:test_id2', 'DELETE:test_id3'])


def test_sqs_trigger(tracer_and_invocation_support, handler, mock_sqs_event, mock_context):
    _, handler = handler
    tracer, invocation_support, invocation_trace_support = tracer_and_invocation_support

    handler(mock_sqs_event, mock_context)
    execution_context = ExecutionContextManager.get()
    span = execution_context.recorder.get_spans()[0]

    assert lambda_event_utils.get_lambda_event_type(mock_sqs_event,
                                                    mock_context) == lambda_event_utils.LambdaEventType.SQS

    assert span.get_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == constants.DomainNames['MESSAGING']
    assert span.get_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == constants.ClassNames['SQS']
    assert span.get_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['MyQueue']

    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == constants.DomainNames[
        'MESSAGING']
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == constants.ClassNames['SQS']
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['MyQueue']

    assert invocation_trace_support.get_incoming_trace_links().get('incomingTraceLinks') == [
        "19dd0b57-b21e-4ac1-bd88-01bbb068cb78"]


def test_sns_trigger(tracer_and_invocation_support, handler, mock_sns_event, mock_context):
    _, handler = handler
    tracer, invocation_support, invocation_trace_support = tracer_and_invocation_support
    handler(mock_sns_event, mock_context)
    execution_context = ExecutionContextManager.get()
    span = execution_context.recorder.get_spans()[0]
    assert lambda_event_utils.get_lambda_event_type(mock_sns_event,
                                                    mock_context) == lambda_event_utils.LambdaEventType.SNS

    assert span.get_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == constants.DomainNames['MESSAGING']
    assert span.get_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == constants.ClassNames['SNS']
    assert span.get_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['ExampleTopic']

    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == constants.DomainNames[
        'MESSAGING']
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == constants.ClassNames['SNS']
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['ExampleTopic']

    assert invocation_trace_support.get_incoming_trace_links().get('incomingTraceLinks') == [
        "95df01b4-ee98-5cb9-9903-4c221d41eb5e"]


def test_kinesis_trigger(tracer_and_invocation_support, handler, mock_kinesis_event, mock_context):
    _, handler = handler
    tracer, invocation_support, invocation_trace_support = tracer_and_invocation_support

    handler(mock_kinesis_event, mock_context)
    execution_context = ExecutionContextManager.get()
    span = execution_context.recorder.get_spans()[0]

    assert lambda_event_utils.get_lambda_event_type(mock_kinesis_event,
                                                    mock_context) == lambda_event_utils.LambdaEventType.Kinesis

    assert span.get_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == constants.DomainNames['STREAM']
    assert span.get_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == constants.ClassNames['KINESIS']
    assert span.get_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['example_stream']

    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == constants.DomainNames[
        'STREAM']
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == constants.ClassNames['KINESIS']
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['example_stream']

    links = ["eu-west-2:example_stream:shardId-000000000000:49545115243490985018280067714973144582180062593244200961"]
    assert invocation_trace_support.get_incoming_trace_links().get('incomingTraceLinks') == links


def test_cloudwatch_schedule_trigger(tracer_and_invocation_support, handler, mock_cloudwatch_schedule_event,
                                     mock_context):
    _, handler = handler
    tracer, invocation_support, _ = tracer_and_invocation_support
    handler(mock_cloudwatch_schedule_event, mock_context)

    execution_context = ExecutionContextManager.get()
    span = execution_context.recorder.get_spans()[0]
    event_type = lambda_event_utils.get_lambda_event_type(mock_cloudwatch_schedule_event,
                                                          mock_context)
    assert event_type == lambda_event_utils.LambdaEventType.CloudWatchSchedule

    assert span.get_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == 'Schedule'
    assert span.get_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == 'AWS-CloudWatch-Schedule'
    assert span.get_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['ExampleRule']

    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == 'Schedule'
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == 'AWS-CloudWatch-Schedule'
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['ExampleRule']


def test_cloudwatch_logs_trigger(tracer_and_invocation_support, handler, mock_cloudwatch_logs_event, mock_context):
    _, handler = handler
    tracer, invocation_support, _ = tracer_and_invocation_support
    handler(mock_cloudwatch_logs_event, mock_context)
    execution_context = ExecutionContextManager.get()
    span = execution_context.recorder.get_spans()[0]

    assert lambda_event_utils.get_lambda_event_type(mock_cloudwatch_logs_event,
                                                    mock_context) == lambda_event_utils.LambdaEventType.CloudWatchLogs

    compressed_data = base64.b64decode(mock_cloudwatch_logs_event['awslogs']['data'])
    decompressed_data = json.loads(str(GzipFile(fileobj=BytesIO(compressed_data)).read(), 'utf-8'))

    assert span.get_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == 'Log'
    assert span.get_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == 'AWS-CloudWatch-Log'
    assert span.get_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == [decompressed_data['logGroup']]

    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == 'Log'
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == 'AWS-CloudWatch-Log'
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == [
        decompressed_data['logGroup']]


def test_cloudfront_event_trigger(tracer_and_invocation_support, handler, mock_cloudfront_event, mock_context):
    _, handler = handler
    tracer, invocation_support, _ = tracer_and_invocation_support
    handler(mock_cloudfront_event, mock_context)
    execution_context = ExecutionContextManager.get()
    span = execution_context.recorder.get_spans()[0]

    assert lambda_event_utils.get_lambda_event_type(mock_cloudfront_event,
                                                    mock_context) == lambda_event_utils.LambdaEventType.CloudFront

    assert span.get_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == 'CDN'
    assert span.get_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == 'AWS-CloudFront'
    assert span.get_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['/test']

    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == 'CDN'
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == 'AWS-CloudFront'
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['/test']


def test_firehose_trigger(tracer_and_invocation_support, handler, mock_firehose_event, mock_context):
    thundra, handler = handler
    tracer, invocation_support, invocation_trace_support = tracer_and_invocation_support
    assert lambda_event_utils.get_lambda_event_type(mock_firehose_event,
                                                    mock_context) == lambda_event_utils.LambdaEventType.Firehose
    handler(mock_firehose_event, mock_context)
    execution_context = ExecutionContextManager.get()
    span = execution_context.recorder.get_spans()[0]

    assert span.get_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == constants.DomainNames['STREAM']
    assert span.get_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == constants.ClassNames['FIREHOSE']
    assert span.get_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['exampleStream']

    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == constants.DomainNames[
        'STREAM']
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == constants.ClassNames[
        'FIREHOSE']
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['exampleStream']

    links = [
        "eu-west-2:exampleStream:1495072948:75c5afa1146857f64e92e6bb6e561ded",
        "eu-west-2:exampleStream:1495072949:75c5afa1146857f64e92e6bb6e561ded",
        "eu-west-2:exampleStream:1495072950:75c5afa1146857f64e92e6bb6e561ded",
    ]
    assert sorted(invocation_trace_support.get_incoming_trace_links().get('incomingTraceLinks')) == sorted(links)


def test_apigateway_proxy_trigger(tracer_and_invocation_support, handler, mock_apigateway_proxy_event, mock_context):
    _, handler = handler
    tracer, invocation_support, _ = tracer_and_invocation_support
    handler(mock_apigateway_proxy_event, mock_context)
    execution_context = ExecutionContextManager.get()
    span = execution_context.recorder.get_spans()[0]

    assert lambda_event_utils.get_lambda_event_type(mock_apigateway_proxy_event,
                                                    mock_context) == lambda_event_utils.LambdaEventType.APIGatewayProxy

    assert span.get_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == 'API'
    assert span.get_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == 'AWS-APIGateway'
    assert span.get_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['/{proxy+}']

    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == 'API'
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == 'AWS-APIGateway'
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['/{proxy+}']


def test_apigateway_trigger(tracer_and_invocation_support, handler, mock_apigateway_event, mock_context):
    _, handler = handler
    tracer, invocation_support, _ = tracer_and_invocation_support
    handler(mock_apigateway_event, mock_context)
    execution_context = ExecutionContextManager.get()
    span = execution_context.recorder.get_spans()[0]

    assert lambda_event_utils.get_lambda_event_type(mock_apigateway_event,
                                                    mock_context) == lambda_event_utils.LambdaEventType.APIGateway

    assert span.get_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == 'API'
    assert span.get_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == 'AWS-APIGateway'
    assert span.get_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == [
        'random.execute-api.us-west-2.amazonaws.com/dev{}']

    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == 'API'
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == 'AWS-APIGateway'
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == [
        'random.execute-api.us-west-2.amazonaws.com/dev{}']


def test_s3_trigger(tracer_and_invocation_support, handler, mock_s3_event, mock_context):
    _, handler = handler
    tracer, invocation_support, invocation_trace_support = tracer_and_invocation_support
    assert lambda_event_utils.get_lambda_event_type(mock_s3_event,
                                                    mock_context) == lambda_event_utils.LambdaEventType.S3
    handler(mock_s3_event, mock_context)
    execution_context = ExecutionContextManager.get()
    span = execution_context.recorder.get_spans()[0]

    assert span.get_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == constants.DomainNames['STORAGE']
    assert span.get_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == constants.ClassNames['S3']
    assert span.get_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['example-bucket']

    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == constants.DomainNames[
        'STORAGE']
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == constants.ClassNames['S3']
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['example-bucket']

    assert invocation_trace_support.get_incoming_trace_links().get('incomingTraceLinks') == ["EXAMPLE123456789"]


def test_lambda_trigger(tracer_and_invocation_support, handler, mock_event, mock_lambda_context):
    thundra, handler = handler
    tracer, invocation_support, invocation_trace_support = tracer_and_invocation_support
    assert lambda_event_utils.get_lambda_event_type(mock_event,
                                                    mock_lambda_context) == lambda_event_utils.LambdaEventType.Lambda
    handler(mock_event, mock_lambda_context)
    execution_context = ExecutionContextManager.get()
    span = execution_context.recorder.get_spans()[0]

    assert span.get_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == constants.DomainNames['API']
    assert span.get_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == constants.ClassNames['LAMBDA']
    assert span.get_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['Sample Context']

    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == constants.DomainNames['API']
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == constants.ClassNames['LAMBDA']
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['Sample Context']

    assert invocation_trace_support.get_incoming_trace_links().get('incomingTraceLinks') == ["aws_request_id"]


def test_eventbridge_trigger(tracer_and_invocation_support, handler, mock_eventbridge_event, mock_context):
    _, handler = handler
    tracer, invocation_support, invocation_trace_support = tracer_and_invocation_support
    assert lambda_event_utils.get_lambda_event_type(mock_eventbridge_event,
                                                    mock_context) == lambda_event_utils.LambdaEventType.EventBridge
    handler(mock_eventbridge_event, mock_context)
    execution_context = ExecutionContextManager.get()
    span = execution_context.recorder.get_spans()[0]

    assert span.get_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == constants.DomainNames['MESSAGING']
    assert span.get_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == constants.ClassNames['EVENTBRIDGE']
    assert span.get_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == ['EC2 Command Status-change Notification']

    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME']) == constants.DomainNames[
        'MESSAGING']
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_CLASS_NAME']) == constants.ClassNames[
        'EVENTBRIDGE']
    assert invocation_support.get_agent_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES']) == [
        'EC2 Command Status-change Notification']

    assert invocation_trace_support.get_incoming_trace_links().get('incomingTraceLinks') == [
        "51c0891d-0e34-45b1-83d6-95db273d1602"]



from collections import OrderedDict

from thundra._version import __version__

DATE_FORMAT = "%Y-%m-%d %H:%M:%S.%f %z"

HOST = "https://api.thundra.io/v1"
PATH = "/monitoring-data"
COMPOSITE_DATA_PATH = "/composite-monitoring-data"

THUNDRA_AGENT_VERSION = __version__
TEST_RUN_EVENTS_DATA_VERSION = "1.0"

DEFAULT_APPLICATION_NAME = 'catchpoint-app'
LAMBDA_APPLICATION_DOMAIN_NAME = 'API'
LAMBDA_APPLICATION_CLASS_NAME = 'AWS-Lambda'
LAMBDA_APPLICATION_PLATFORM = 'AWS Lambda'

TRIGGER_OPERATION_NAME_KEY = 'x-catchpoint-trigger-operation-name'
TRIGGER_CLASS_NAME_KEY = 'x-catchpoint-trigger-class-name'
TRIGGER_DOMAIN_NAME_KEY = 'x-catchpoint-trigger-domain-name'
TRIGGER_RESOURCE_NAME_KEY = 'x-catchpoint-resource-name'

THUNDRA_TRACE_ID_KEY = 'x-catchpoint-trace-id'
THUNDRA_TRANSACTION_ID_KEY = 'x-catchpoint-transaction-id'
THUNDRA_SPAN_ID_KEY = 'x-catchpoint-span-id'
THUNDRA_BAGGAGE_PREFIX = 'x-catchpoint-baggage-'

DEFAULT_METRIC_SAMPLING_TIME_FREQ = 5 * 60 * 1000
DEFAULT_METRIC_SAMPLING_COUNT_FREQ = 100

AWS_LAMBDA_APPLICATION_ID = 'AWS_LAMBDA_APPLICATION_ID'
AWS_SAM_LOCAL = 'AWS_SAM_LOCAL'
AWS_LAMBDA_LOG_STREAM_NAME = 'AWS_LAMBDA_LOG_STREAM_NAME'
AWS_LAMBDA_FUNCTION_VERSION = 'AWS_LAMBDA_FUNCTION_VERSION'
AWS_LAMBDA_FUNCTION_NAME = 'AWS_LAMBDA_FUNCTION_NAME'
AWS_REGION = 'AWS_REGION'
AWS_LAMBDA_FUNCTION_MEMORY_SIZE = "AWS_LAMBDA_FUNCTION_MEMORY_SIZE"
AWS_LAMBDA_APPLICATION_DOMAIN_NAME = 'API'
AWS_LAMBDA_APPLICATION_CLASS_NAME = 'AWS-Lambda'
LOCAL_COLLECTOR_ENDPOINT = 'localhost:3333'

CONTEXT_FUNCTION_NAME = 'function_name'
CONTEXT_FUNCTION_VERSION = 'function_version'
CONTEXT_APPLICATION_PLATFORM = 'AWS-Lambda'
CONTEXT_MEMORY_LIMIT_IN_MB = 'memory_limit_in_mb'
CONTEXT_INVOKED_FUNCTION_ARN = 'invoked_function_arn'
CONTEXT_AWS_REQUEST_ID = 'aws_request_id'
CONTEXT_LOG_GROUP_NAME = 'log_group_name'
CONTEXT_LOG_STREAM_NAME = 'log_stream_name'

TRACE_ARGS = 'trace_args'
TRACE_RETURN_VALUE = 'trace_return_value'
TRACE_ERROR = 'trace_error'

MAX_INCOMING_TRACE_LINKS = 10
MAX_OUTGOING_TRACE_LINKS = 20

LIST_SEPARATOR = ','

DEFAULT_LAMBDA_TIMEOUT_MARGIN = 200
DATA_FORMAT_VERSION = '2.0'

THUNDRA_LAMBDA_SPAN_LISTENER_INFO_TAG = 'thundra.span_listener.info'

DEFAULT_REPORT_REST_COMPOSITE_BATCH_SIZE = 100
DEFAULT_REPORT_CLOUDWATCH_COMPOSITE_BATCH_SIZE = 10

#### INTEGRATIONS ####

THUNDRA_MAX_STREAM_REQUEST_BODY = 200 * 1024
DEFAULT_MONGO_COMMAND_SIZE_LIMIT = 128 * 1024
DEFAULT_REPORT_TIMEOUT = 5

AWS_SERVICE_REQUEST = 'AWSServiceRequest'

LineByLineTracingTags = {
    'lines': 'method.lines',
    'next_span_ids': 'nextSpanIds',
    'source': 'method.source',
    'start_line': 'method.startLine',
    'args': 'method.args'
}

DomainNames = {
    'AWS': 'AWS',
    'DB': 'DB',
    'MESSAGING': 'Messaging',
    'STREAM': 'Stream',
    'STORAGE': 'Storage',
    'API': 'API',
    'CACHE': 'Cache',
    'SCHEDULE': 'Schedule',
    'LOG': 'Log',
    'CDN': 'CDN',
}

ClassNames = {
    'AWSSERVICE': 'AWSService',
    'DYNAMODB': 'AWS-DynamoDB',
    'SQS': 'AWS-SQS',
    'SNS': 'AWS-SNS',
    'KINESIS': 'AWS-Kinesis',
    'FIREHOSE': 'AWS-Firehose',
    'S3': 'AWS-S3',
    'LAMBDA': 'AWS-Lambda',
    'RDB': 'RDB',
    'REDIS': 'Redis',
    'HTTP': 'HTTP',
    'MYSQL': 'MYSQL',
    'POSTGRESQL': 'POSTGRESQL',
    'ELASTICSEARCH': 'ELASTICSEARCH',
    'SCHEDULE': 'AWS-CloudWatch-Schedule',
    'CLOUDWATCHLOG': 'AWS-CloudWatch-Log',
    'CLOUDFRONT': 'AWS-CloudFront',
    'APIGATEWAY': 'AWS-APIGateway',
    'MONGODB': 'MONGODB',
    'SQLALCHEMY': 'SQLALCHEMY',
    'SQLITE': 'SQLITE',
    'ATHENA': 'AWS-Athena',
    'STEPFUNCTIONS': 'AWS-StepFunctions',
    'EVENTBRIDGE': 'AWS-EventBridge',
    'SES': 'AWS-SES',
    'DJANGO': 'Django',
    'FLASK': 'Flask',
    "FASTAPI": "Fastapi",
    "TORNADO": "Tornado"
}

DBTags = {
    'DB_STATEMENT': 'db.statement',
    'DB_STATEMENT_TYPE': 'db.statement.type',
    'DB_INSTANCE': 'db.instance',
    'DB_TYPE': 'db.type',
    'DB_HOST': 'db.host',
    'DB_PORT': 'db.port',
    'DB_USER': 'db.user',
}

AwsDynamoTags = {
    'TABLE_NAME': 'aws.dynamodb.table.name',
    'REQUEST_THROTTLED': 'aws.dynamodb.request.throttled',
}

DBTypes = {
    'DYNAMODB': 'aws-dynamodb',
    'REDIS': 'redis',
    'ELASTICSEARCH': 'elasticsearch',
}

SpanTags = {
    'OPERATION_TYPE': 'operation.type',
    'DB_INSTANCE': 'db.instance',
    'DB_TYPE': 'db.type',
    'DB_HOST': 'db.host',
    'TRIGGER_DOMAIN_NAME': 'trigger.domainName',
    'TRIGGER_CLASS_NAME': 'trigger.className',
    'TRIGGER_OPERATION_NAMES': 'trigger.operationNames',
    'TOPOLOGY_VERTEX': 'topology.vertex',
    'DB_STATEMENT': 'db.statement',
    'DB_STATEMENT_TYPE': 'db.statement.type',
    'TRACE_LINKS': 'trace.links',
    'RESOURCE_NAMES': 'resource.names',
    'RESOURCE_TRACE_LINKS': 'resource.trace.links'
}

SecurityTags = {
    'BLOCKED': 'security.blocked',
    'VIOLATED': 'security.violated'
}

AthenaTags = {
    'S3_OUTPUT_LOCATION': "aws.athena.s3.outputLocation",
    'REQUEST_QUERY_EXECUTION_IDS': "aws.athena.request.query.executionIds",
    'RESPONSE_QUERY_EXECUTION_IDS': "aws.athena.response.query.executionIds",
    'REQUEST_NAMED_QUERY_IDS': "aws.athena.request.namedQuery.ids",
    'RESPONSE_NAMED_QUERY_IDS': "aws.athena.response.namedQuery.ids",
}

AwsSDKTags = {
    'SERVICE_NAME': 'aws.service.name',
    'REQUEST_NAME': 'aws.request.name',
    'HOST': 'host',
}

AwsStepFunctionsTags = {
    'STATE_MACHINE_ARN': 'aws.sf.state_machine.arn',
    'EXECUTION_NAME': 'aws.sf.execution.name',
    'EXECUTION_INPUT': 'aws.sf.execution.input',
    'EXECUTION_ARN': 'aws.sf.execution.arn',
    'EXECUTION_START_DATE': 'aws.sf.execution.start_date'
}

AwsSQSTags = {
    'QUEUE_NAME': 'aws.sqs.queue.name',
    'MESSAGE': 'aws.sqs.message',
    'MESSAGES': 'aws.sqs.messages',
}

AwsSNSTags = {
    'TOPIC_NAME': 'aws.sns.topic.name',
    'MESSAGE': 'aws.sns.message',
}

AwsKinesisTags = {
    'STREAM_NAME': 'aws.kinesis.stream.name',
}

AwsFirehoseTags = {
    'STREAM_NAME': 'aws.firehose.stream.name',
}

AwsS3Tags = {
    'BUCKET_NAME': 'aws.s3.bucket.name',
    'OBJECT_NAME': 'aws.s3.object.name',
}

AwsSESTags = {
    'SERVICE_REQUEST': 'AWSSESRequest',
    'SUBJECT': 'aws.ses.mail.subject',
    'BODY': 'aws.ses.mail.body',
    'TEMPLATE_NAME': 'aws.ses.mail.template.name',
    'TEMPLATE_ARN': 'aws.ses.mail.template.arn',
    'TEMPLATE_DATA': 'aws.ses.mail.template.data',
    'SOURCE': 'aws.ses.mail.source',
    'DESTINATION': 'aws.ses.mail.destination',
}

AwsLambdaTags = {
    'FUNCTION_NAME': 'aws.lambda.name',
    'FUNCTION_QUALIFIER': 'aws.lambda.qualifier',
    'INVOCATION_TYPE': 'aws.lambda.invocation.type',
    'INVOCATION_PAYLOAD': 'aws.lambda.invocation.payload',
}

LambdaRequestType = {
    'InvokeAsync': 'CALL',
    'Invoke': 'CALL',
}

HttpTags = {
    'HTTP_METHOD': 'http.method',
    'HTTP_URL': 'http.url',
    'HTTP_PATH': 'http.path',
    'HTTP_HOST': 'http.host',
    'HTTP_STATUS': 'http.status_code',
    'QUERY_PARAMS': 'http.query_params',
    'BODY': 'http.body',
}

RedisCommandTypes = {
    'APPEND': 'WRITE',
    'BGREWRITEAOF': 'WRITE',
    'BGSAVE': 'WRITE',
    'BITCOUNT': 'READ',
    'BITFIELD': 'WRITE',
    'BITOP': 'WRITE',
    'BITPOS': 'READ',
    'BLPOP': 'DELETE',
    'BRPOP': 'DELETE',
    'BRPOPLPUSH': 'WRITE',
    'BZPOPMIN': 'DELETE',
    'BZPOPMAX': 'DELETE',
    'DBSIZE': 'READ',
    'DECR': 'WRITE',
    'DECRBY': 'WRITE',
    'DELETE': 'DELETE',
    'EVAL': 'EXECUTE',
    'EVALSHA': 'EXECUTE',
    'EXISTS': 'READ',
    'EXPIRE': 'WRITE',
    'EXPIREAT': 'WRITE',
    'FLUSHALL': 'DELETE',
    'FLUSHDB': 'DELETE',
    'GEOADD': 'WRITE',
    'GEOHASH': 'READ',
    'GEOPOS': 'READ',
    'GEODIST': 'READ',
    'GEORADIUS': 'READ',
    'GEORADIUSBYMEMBER': 'READ',
    'GET': 'READ',
    'GETBIT': 'READ',
    'GETRANGE': 'READ',
    'GETSET': 'WRITE',
    'HDEL': 'DELETE',
    'HEXISTS': 'READ',
    'HGET': 'READ',
    'HGETALL': 'READ',
    'HINCRBY': 'WRITE',
    'HINCRBYFLOAT': 'WRITE',
    'HKEYS': 'READ',
    'HLEN': 'READ',
    'HMGET': 'READ',
    'HMSET': 'WRITE',
    'HSET': 'WRITE',
    'HSETNX': 'WRITE',
    'HSTRLEN': 'READ',
    'HVALS': 'READ',
    'INCR': 'WRITE',
    'INCRBY': 'WRITE',
    'INCRBYFLOAT': 'WRITE',
    'KEYS': 'READ',
    'LINDEX': 'READ',
    'LINSERT': 'WRITE',
    'LLEN': 'READ',
    'LPOP': 'DELETE',
    'LPUSH': 'WRITE',
    'LPUSHX': 'WRITE',
    'LRANGE': 'READ',
    'LREM': 'DELETE',
    'LSET': 'WRITE',
    'LTRIM': 'DELETE',
    'MGET': 'READ',
    'MSET': 'WRITE',
    'MSETNX': 'WRITE',
    'PERSIST': 'WRITE',
    'PEXPIRE': 'WRITE',
    'PEXPIREAT': 'WRITE',
    'PFADD': 'WRITE',
    'PFCOUNT': 'READ',
    'PFMERGE': 'WRITE',
    'PSETEX': 'WRITE',
    'PUBLISH': 'WRITE',
    'RPOP': 'DELETE',
    'RPOPLPUSH': 'WRITE',
    'RPUSH': 'WRITE',
    'RPUSHX': 'WRITE',
    'SADD': 'WRITE',
    'SCARD': 'READ',
    'SDIFFSTORE': 'WRITE',
    'SET': 'WRITE',
    'SETBIT': 'WRITE',
    'SETEX': 'WRITE',
    'SETNX': 'WRITE',
    'SETRANGE': 'WRITE',
    'SINTER': 'READ',
    'SINTERSTORE': 'WRITE',
    'SISMEMBER': 'READ',
    'SMEMBERS': 'READ',
    'SMOVE': 'WRITE',
    'SORT': 'WRITE',
    'SPOP': 'DELETE',
    'SRANDMEMBER': 'READ',
    'SREM': 'DELETE',
    'STRLEN': 'READ',
    'SUNION': 'READ',
    'SUNIONSTORE': 'WRITE',
    'ZADD': 'WRITE',
    'ZCARD': 'READ',
    'ZCOUNT': 'READ',
    'ZINCRBY': 'WRITE',
    'ZINTERSTORE': 'WRITE',
    'ZLEXCOUNT': 'READ',
    'ZPOPMAX': 'DELETE',
    'ZPOPMIN': 'DELETE',
    'ZRANGE': 'READ',
    'ZRANGEBYLEX': 'READ',
    'ZREVRANGEBYLEX': 'READ',
    'ZRANGEBYSCORE': 'READ',
    'ZRANK': 'READ',
    'ZREM': 'DELETE',
    'ZREMRANGEBYLEX': 'DELETE',
    'ZREMRANGEBYRANK': 'DELETE',
    'ZREMRANGEBYSCORE': 'DELETE',
    'ZREVRANGE': 'READ',
    'ZREVRANGEBYSCORE': 'READ',
    'ZREVRANK': 'READ',
    'ZSCORE': 'READ',
    'ZUNIONSTORE': 'WRITE',
    'SCAN': 'READ',
    'SSCAN': 'READ',
    'HSCAN': 'READ',
    'ZSCAN': 'READ',
    'XADD': 'WRITE',
    'XRANGE': 'READ',
    'XREVRANGE': 'READ',
    'XLEN': 'READ',
    'XREAD': 'READ',
    'XREADGROUP': 'READ',
    'XPENDING': 'READ',
}

RedisTags = {
    'REDIS_HOST': 'redis.host',
    'REDIS_PORT': 'redis.port',
    'REDIS_COMMAND': 'redis.command',
    'REDIS_COMMANDS': 'redis.commands',
    'REDIS_COMMAND_TYPE': 'redis.command.type',
    'REDIS_COMMAND_ARGS': 'redis.command.args',
}

ESTags = {
    'ES_URI': 'elasticsearch.uri',
    'ES_METHOD': 'elasticsearch.method',
    'ES_PARAMS': 'elasticsearch.params',
    'ES_BODY': 'elasticsearch.body',
    'ES_HOSTS': 'elasticsearch.hosts',
    'ES_NORMALIZED_URI': 'elasticsearch.normalized_uri'
}

AwsXrayConstants = {
    'DEFAULT_OPERATION_NAME': 'AWS X-Ray',
    'XRAY_SUBSEGMENTED_TAG_NAME': 'THUNDRA::XRAY_SUBSEGMENTED',
}

MongoDBTags = {
    'MONGODB_COMMAND': 'mongodb.command',
    'MONGODB_COMMAND_NAME': 'mongodb.command.name',
    'MONGODB_COLLECTION': 'mongodb.collection.name',
}

MongoDBCommandTypes = {
    # Aggregate Commands
    'AGGREGATE': 'READ',
    'COUNT': 'READ',
    'DISTINCT': 'READ',
    'GROUP': 'READ',
    'MAPREDUCE': 'READ',

    # Geospatial Commands
    'GEONEAR': 'READ',
    'GEOSEARCH': 'READ',

    # Query and Write Operation Commands
    'DELETE': 'DELETE',
    'EVAL': 'EXECUTE',
    'FIND': 'READ',
    'FINDANDMODIFY': 'WRITE',
    'GETLASTERROR': 'READ',
    'GETMORE': 'READ',
    'GETPREVERROR': 'READ',
    'INSERT': 'WRITE',
    'PARALLELCOLLECTIONSCAN': 'READ',
    'RESETERROR': 'WRITE',
    'UPDATE': 'WRITE',

    # Query Plan Cache Commands
    'PLANCACHECLEAR': 'DELETE',
    'PLANCACHECLEARFILTERS': 'DELETE',
    'PLANCACHELISTFILTERS': 'READ',
    'PLANCACHELISTPLANS': 'READ',
    'PLANCACHELISTQUERYSHAPES': 'READ',
    'PLANCACHESETFILTER': 'WRITE',

    # Authentication Commands
    'AUTHENTICATE': 'EXECUTE',
    'LOGOUT': 'EXECUTE',

    # User Management Commands
    'CREATEUSER': 'WRITE',
    'DROPALLUSERSFROMDATABASE': 'DELETE',
    'DROPUSER': 'DELETE',
    'GRANROLESTOUSER': 'WRITE',
    'REVOKEROLESFROMUSER': 'WRITE',
    'UPDATEUSER': 'WRITE',
    'USERSINFO': 'READ',

    # Role Management Commands
    'CREATEROLE': 'WRITE',
    'DROPROLE': 'DELETE',
    'DROPALLROLESFROMDATABASE': 'DELETE',
    'GRANTPRIVILEGESTOROLE': 'WRITE',
    'GRANTROLESTOROLE': 'WRITE',
    'INVALIDATEUSERCACHE': 'DELETE',
    'REVOKEPRIVILEGESFROMROLE': 'WRITE',
    'REVOKEROLESFROMROLE': 'WRITE',
    'ROLESINFO': 'READ',
    'UPDATEROLE': 'WRITE',

    # Replication Commands
    'ISMASTER': 'READ',
    'REPLSETABORTPRIMARYCATCHUP': 'EXECUTE',
    'REPLSETFREEZE': 'EXECUTE',
    'REPLSETGETCONFIG': 'READ',
    'REPLSETGETSTATUS': 'READ',
    'REPLSETINITIATE': 'EXECUTE',
    'REPLSETMAINTENANCE': 'EXECUTE',
    'REPLSETRECONFIG': 'EXECUTE',
    'REPLSETRESIZEOPLOG': 'EXECUTE',
    'REPLSETSTEPDOWN': 'EXECUTE',
    'REPLSETSYNCFROM': 'EXECUTE',

    # Sharding Commands
    'ADDSHARD': 'EXECUTE',
    'ADDSHARDTOZONE': 'EXECUTE',
    'BALANCERSTART': 'EXECUTE',
    'BALANCERSTATUS': 'READ',
    'BALANCERSTOP': 'EXECUTE',
    'CLEANUPORPHANED': 'EXECUTE',
    'ENABLESHARDING': 'EXECUTE',
    'FLUSHROUTERCONFIG': 'EXECUTE',
    'ISDBGRID': 'READ',
    'LISTSHARDS': 'READ',
    'MOVEPRIMARY': 'EXECUTE',
    'MERGECHUNKS': 'EXECUTE',
    'REMOVESHARD': 'EXECUTE',
    'REMOVESHARDFROMZONE': 'EXECUTE',
    'SHARDCOLLECTION': 'EXECUTE',
    'SHARDINGSTATE': 'READ',
    'SPLIT': 'EXECUTE',
    'UPDATEZONEKEYRANGE': 'EXECUTE',

    # Session Commands
    'ABORTTRANSACTION': 'EXECUTE',
    'COMMITTRANSACTION': 'EXECUTE',
    'ENDSESSIONS': 'EXECUTE',
    'KILLALLSESSIONS': 'EXECUTE',
    'KILLALLSESSIONSBYPATTERN': 'EXECUTE',
    'KILLSESSIONS': 'EXECUTE',
    'REFRESHSESSIONS': 'EXECUTE',
    'STARTSESSION': 'EXECUTE',

    # Administration Commands 
    'CLONE': 'EXECUTE',
    'CLONECOLLECTION': 'EXECUTE',
    'CLONECOLLECTIONASCAPPED': 'EXECUTE',
    'COLLMOD': 'WRITE',
    'COMPACT': 'EXECUTE',
    'CONVERTTOCAPPED': 'EXECUTE',
    'COPYDB': 'EXECUTE',
    'CREATE': 'WRITE',
    'CREATEINDEXES': 'WRITE',
    'CURRENTOP': 'READ',
    'DROP': 'DELETE',
    'DROPDATABASE': 'DELETE',
    'DROPINDEXES': 'DELETE',
    'FILEMD5': 'READ',
    'FSYNC': 'EXECUTE',
    'FSYNCUNLOCK': 'EXECUTE',
    'GETPARAMETER': 'READ',
    'KILLCURSORS': 'EXECUTE',
    'KILLOP': 'EXECUTE',
    'LISTCOLLECTIONS': 'READ',
    'LISTDATABASES': 'READ',
    'LISTINDEXES': 'READ',
    'LOGROTATE': 'EXECUTE',
    'REINDEX': 'WRITE',
    'RENAMECOLLECTION': 'WRITE',
    'REPAIRDATABASE': 'EXECUTE',
    'SETFEATURECOMPATIBILITYVERSION': 'WRITE',
    'SETPARAMETER': 'WRITE',
    'SHUTDOWN': 'EXECUTE',
    'TOUCH': 'EXECUTE',

    # Diagnostic Commands
    'BUILDINFO': 'READ',
    'COLLSTATS': 'READ',
    'CONNPOOLSTATS': 'READ',
    'CONNECTIONSTATUS': 'READ',
    'CURSORINFO': 'READ',
    'DBHASH': 'READ',
    'DBSTATS': 'READ',
    'DIAGLOGGING': 'READ',
    'EXPLAIN': 'READ',
    'FEATURES': 'READ',
    'GETCMDLINEOPTS': 'READ',
    'GETLOG': 'READ',
    'HOSTINFO': 'READ',
    'LISTCOMMANDS': 'READ',
    'PROFILE': 'READ',
    'SERVERSTATUS': 'READ',
    'SHARDCONNPOOLSTATS': 'READ',
    'TOP': 'READ',

    # Free Monitoring Commands
    'SETFREEMONITORING': 'EXECUTE',

    # Auditing Commands
    'LOGAPPLICATIONMESSAGE': 'EXECUTE',
}

OperationTypeMappings = {
    'exclusions': {
        'AWS-Lambda': {
            'ListTags': 'READ',
            'TagResource': 'WRITE',
            'UntagResource': 'WRITE',
            'EnableReplication': 'PERMISSION'
        },
        'AWS-S3': {
            'HeadBucket': 'LIST',
            'ListBucketByTags': 'READ',
            'ListBucketMultipartUploads': 'READ',
            'ListBucketVersions': 'READ',
            'ListJobs': 'READ',
            'ListMultipartUploadParts': 'READ',
            'GetBucketTagging': 'READ',
            'GetObjectVersionTagging': 'READ',
            'GetObjectTagging': 'READ',
            'GetBucketObjectLockConfiguration': 'WRITE',
            'GetObjectLegalHold': 'WRITE',
            'GetObjectRetention': 'WRITE',
            'DeleteObjectTagging': 'TAGGING',
            'DeleteObjectVersionTagging': 'TAGGING',
            'PutBucketTagging': 'TAGGING',
            'PutObjectTagging': 'TAGGING',
            'PutObjectVersionTagging': 'TAGGING',
            'AbortMultipartUpload': 'WRITE',
            'ReplicateDelete': 'WRITE',
            'ReplicateObject': 'WRITE',
            'RestoreObject': 'WRITE',
            'DeleteBucketPolicy': 'PERMISSION',
            'ObjectOwnerOverrideToBucketOwner': 'PERMISSION',
            'PutAccountPublicAccessBlock': 'PERMISSION',
            'PutBucketAcl': 'PERMISSION',
            'PutBucketPolicy': 'PERMISSION',
            'PutBucketPublicAccessBlock': 'PERMISSION',
            'PutObjectAcl': 'PERMISSION',
            'PutObjectVersionAcl': 'PERMISSION'
        },
        'AWS-SNS': {
            'ListPhoneNumbersOptedOut': 'READ',
            'ListTagsForResource': 'READ',
            'CheckIfPhoneNumberIsOptedOut': 'READ',
            'UntagResource': 'TAGGING',
            'ConfirmSubscription': 'WRITE',
            'OptInPhoneNumber': 'WRITE',
            'Subscribe': 'WRITE',
            'Unsubscribe': 'WRITE'
        },
        'AWS-Athena': {
            'BatchGetNamedQuery': 'READ',
            'BatchGetQueryExecution': 'READ',
            'ListTagsForResource': 'LIST',
            'CreateWorkGroup': 'WRITE',
            'UntagResource': 'TAGGING',
            'TagResource': 'TAGGING',
            'CancelQueryExecution': 'WRITE',
            'RunQuery': 'WRITE',
            'StartQueryExecution': 'WRITE',
            'StopQueryExecution': 'WRITE'
        },
        'AWS-Kinesis': {
            'ListTagsForStream': 'READ',
            'SubscribeToShard': 'READ',
            'AddTagsToStream': 'TAGGING',
            'RemoveTagsFromStream': 'TAGGING',
            'DecreaseStreamRetentionPeriod': 'WRTITE',
            'DeregisterStreamConsumer': 'WRITE',
            'DisableEnhancedMonitoring': 'WRITE',
            'EnableEnhancedMonitoring': 'WRITE',
            'IncreaseStreamRetentionPeriod': 'WRITE',
            'MergeShards': 'WRITE',
            'RegisterStreamConsumer': 'WRITE',
            'SplitShard': 'WRITE',
            'UpdateShardCount': 'WRITE'
        },
        'AWS-Firehose': {
            'DescribeDeliveryStream': 'LIST',
            'StartDeliveryStreamEncryption': 'WRITE',
            'StopDeliveryStreamEncryption': 'WRITE',
            'TagDeliveryStream': 'WRITE',
            'UntagDeliveryStream': 'WRITE'
        },
        'AWS-SQS': {
            'ListDeadLetterSourceQueues': 'READ',
            'ListQueueTags': 'READ',
            'ReceiveMessage': 'READ',
            'TagQueue': 'TAGGING',
            'UntagQueue': 'TAGGING',
            'PurgeQueue': 'WRITE',
            'SetQueueAttributes': 'WRITE'
        },
        'AWS-DynamoDB': {
            'BatchGetItem': 'READ',
            'ConditionCheckItem': 'READ',
            'ListStreams': 'READ',
            'ListTagsOfResource': 'READ',
            'Query': 'READ',
            'Scan': 'READ',
            'TagResource': 'TAGGING',
            'UntagResource': 'TAGGING',
            'BatchWriteItem': 'WRITE',
            'PurchaseReservedCapacityOfferings': 'WRITE',
            'RestoreTableFromBackup': 'WRITE',
            'RestoreTableToPointInTime': 'WRITE'
        },
        'AWS-EventBridge': {
            'TestEventPattern': 'READ',
            'PutRule': 'TAGGING',
            'ActivateEventSource': 'WRITE',
            'DeactivateEventSource': 'WRITE',
            'DisableRule': 'WRITE',
            'EnableRule': 'WRITE',
            'PutEvents': 'WRITE',
            'PutPartnerEvents': 'WRITE',
            'PutPermission': 'WRITE',
            'PutTargets': 'WRITE',
            'RemovePermission': 'WRITE',
            'RemoveTargets': 'WRITE',
        },
        'AWS-SES': {
            'VerifyDomainDkim': 'READ',
            'VerifyDomainIdentity': 'READ',
            'VerifyEmailAddress': 'READ',
            'VerifyEmailIdentity': 'READ',
            'CloneReceiptRuleSet': 'WRITE',
            'ReorderReceiptRuleSet': 'WRITE',
            'TestRenderTemplate': 'WRITE',
        },
        'AWS-StepFunctions': {
            'StartExecution': 'EXECUTE'
        }
    },
    'patterns': OrderedDict([
        ('^List.*$', 'LIST'),
        ('^Get.*$', 'READ'),
        ('^Create.*$', 'WRITE'),
        ('^Delete.*$', 'WRITE'),
        ('^Invoke.*$', 'WRITE'),
        ('^Publish.*$', 'WRITE'),
        ('^Put.*$', 'WRITE'),
        ('^Update.*$', 'WRITE'),
        ('^Describe.*$', 'READ'),
        ('^Change.*$', 'WRITE'),
        ('^Send.*$', 'WRITE'),
        ('^.*Permission$', 'PERMISSION'),
        ('^.*Tagging$', 'TAGGING'),
        ('^.*Tags$', 'TAGGING'),
        ('^Set.*$', 'WRITE')
    ])
}

AwsEventBridgeTags = {
    'SERVICE_REQUEST': 'AWSEventBridgeRequest',
    'ENTRIES': 'aws.eventbridge.entries'
}

CatchpointTags = {
    'REGION_NAME': 'catchpoint.region.name',
    'COUNTRY_NAME': 'catchpoint.country.name',
    'CITY_NAME': 'catchpoint.city.name',
    'TEST_ID': 'catchpoint.test.id'
}

HTTPHeaders = {
    'USER_AGENT': 'User-Agent',
    'CONTENT_TYPE': 'Content-Type',
    'AUTHORIZATION': 'Authorization'
}

CatchpointHeaders = {
    'REGION_NAME': 'x-catchpoint-region-name',
    'COUNTRY_NAME': 'x-catchpoint-country-name',
    'CITY_NAME': 'x-catchpoint-city-name',
    'TEST_ID': 'x-catchpoint-test-id',
    'TIME': 'x-catchpoint-time'
}

CatchpointProperties = {
    'DEFAULT_APP_NAME': 'Catchpoint',
    'APP_DOMAIN_NAME': 'Monitoring',
    'APP_CLASS_NAME': 'Catchpoint',
    'APP_REGION_PLACEHOLDER': '${cp-app-region}',
    'APP_NAME_PLACEHOLDER': '${cp-app-name}',
    'APP_ID_TEMPLATE': ':catchpoint:${cp-app-region}:${cp-app-name}',
    'HTTP_REQUEST_DOMAIN_NAME': 'API',
    'HTTP_REQUEST_CLASS_NAME': 'HTTP',
    'TIME_MAX_DIFF': 30000,
    'AGENT_VERSION': 'Catchpoint'
}





import thundra.constants as constants
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.integrations.base_integration import BaseIntegration
from thundra.integrations.rdb_base import RdbBaseIntegration


class DjangoORMIntegration(BaseIntegration, RdbBaseIntegration):
    CLASS_TYPE = 'django_orm'

    def __init__(self):
        pass

    def get_operation_name(self, wrapped, instance, args, kwargs):
        try:
            host = args[3]['cursor'].db.settings_dict.get('HOST')
        except:
            return ''
        return host

    def before_call(self, scope, execute, instance, _args, _kwargs, response, exception):
        scope.span.domain_name = constants.DomainNames['DB']
        scope.span.class_name = constants.ClassNames['RDB']

        try:
            db = _args[3]['cursor'].db
            settings = db.settings_dict
            vendor = db.vendor.upper()
            scope.span.class_name = constants.ClassNames.get(vendor, constants.ClassNames['RDB'])
        except:
            settings = {}
            vendor = ''

        query = ''
        operation = ''
        try:
            query = _args[0]
            if len(query) > 0:
                operation = query.split()[0].strip("\"").lower()
        except:
            pass

        tags = {
            constants.SpanTags['OPERATION_TYPE']: DjangoORMIntegration._OPERATION_TO_TYPE.get(operation, ''),
            constants.SpanTags['DB_INSTANCE']: settings.get('NAME'),
            constants.SpanTags['DB_HOST']: settings.get('HOST'),
            constants.SpanTags['DB_TYPE']: vendor,
            constants.SpanTags['DB_STATEMENT_TYPE']: operation.upper(),
            constants.SpanTags['TOPOLOGY_VERTEX']: True,
        }

        if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_RDB_STATEMENT_MASK):
            tags[constants.DBTags['DB_STATEMENT']] = query

        scope.span.tags = tags



from thundra.config import config_names

CONFIG_METADATA = {
    config_names.THUNDRA_APIKEY: {
        'type': 'string',
    },
    config_names.THUNDRA_DEBUG_ENABLE: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_DISABLE: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_DISABLE: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_METRIC_DISABLE: {
        'type': 'boolean',
        'defaultValue': True,
    },
    config_names.THUNDRA_LOG_DISABLE: {
        'type': 'boolean',
        'defaultValue': True,
    },
    config_names.THUNDRA_APPLICATION_ID: {
        'type': 'string',
    },
    config_names.THUNDRA_APPLICATION_INSTANCE_ID: {
        'type': 'string',
    },
    config_names.THUNDRA_APPLICATION_NAME: {
        'type': 'string',
    },
    config_names.THUNDRA_APPLICATION_STAGE: {
        'type': 'string',
    },
    config_names.THUNDRA_APPLICATION_DOMAIN_NAME: {
        'type': 'string',
        'defaultValue': 'API',
    },
    config_names.THUNDRA_APPLICATION_CLASS_NAME: {
        'type': 'string',
        'defaultValue': 'AWS-Lambda',
    },
    config_names.THUNDRA_APPLICATION_VERSION: {
        'type': 'string',
    },
    config_names.THUNDRA_APPLICATION_TAG_PREFIX: {
        'type': 'any',
    },
    config_names.THUNDRA_REPORT_REST_BASEURL: {
        'type': 'string',
        'defaultValue': 'https://collector.thundra.io/v1',
    },
    config_names.THUNDRA_REPORT_CLOUDWATCH_ENABLE: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_REPORT_REST_LOCAL: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_REPORT_REST_COMPOSITE_BATCH_SIZE: {
        'type': 'int',
        'defaultValue': 100,
    },
    config_names.THUNDRA_REPORT_CLOUDWATCH_COMPOSITE_BATCH_SIZE: {
        'type': 'int',
        'defaultValue': 10,
    },
    config_names.THUNDRA_REPORT_REST_COMPOSITE_ENABLE: {
        'type': 'boolean',
        'defaultValue': True,
    },
    config_names.THUNDRA_REPORT_CLOUDWATCH_COMPOSITE_ENABLE: {
        'type': 'boolean',
        'defaultValue': True,
    },
    config_names.THUNDRA_LAMBDA_HANDLER: {
        'type': 'string',
    },
    config_names.THUNDRA_LAMBDA_WARMUP_WARMUPAWARE: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_LAMBDA_TIMEOUT_MARGIN: {
        'type': 'int',
    },
    config_names.THUNDRA_LAMBDA_ERROR_STACKTRACE_MASK: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_REQUEST_SKIP: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_RESPONSE_SKIP: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_LAMBDA_TRACE_KINESIS_REQUEST_ENABLE: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_LAMBDA_TRACE_FIREHOSE_REQUEST_ENABLE: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_LAMBDA_TRACE_CLOUDWATCHLOG_REQUEST_ENABLE: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_LAMBDA_AWS_STEPFUNCTIONS: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_LAMBDA_AWS_APPSYNC: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_INSTRUMENT_DISABLE: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_INSTRUMENT_TRACEABLECONFIG: {
        'type': 'string',
    },
    config_names.THUNDRA_TRACE_SPAN_LISTENERCONFIG: {
        'type': 'string',
    },
    config_names.THUNDRA_SAMPLER_TIMEAWARE_TIMEFREQ: {
        'type': 'int',
        'defaultValue': 300000,
    },
    config_names.THUNDRA_SAMPLER_COUNTAWARE_COUNTFREQ: {
        'type': 'int',
        'defaultValue': 100
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_AWS_SNS_MESSAGE_MASK: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_AWS_SNS_TRACEINJECTION_DISABLE: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_AWS_SQS_MESSAGE_MASK: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_AWS_SQS_TRACEINJECTION_DISABLE: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_AWS_LAMBDA_PAYLOAD_MASK: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_AWS_LAMBDA_TRACEINJECTION_DISABLE: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_AWS_DYNAMODB_STATEMENT_MASK: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_AWS_DYNAMODB_TRACEINJECTION_ENABLE: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_AWS_ATHENA_STATEMENT_MASK: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_HTTP_BODY_MASK: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_HTTP_URL_DEPTH: {
        'type': 'int',
        'defaultValue': 1,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_HTTP_TRACEINJECTION_DISABLE: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_HTTP_ERROR_STATUS_CODE_MIN: {
        'type': 'int',
        'defaultValue': 400,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_REDIS_COMMAND_MASK: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_RDB_STATEMENT_MASK: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_ELASTICSEARCH_BODY_MASK: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_ELASTICSEARCH_PATH_DEPTH: {
        'type': 'int',
        'defaultValue': 1,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_MONGODB_COMMAND_MASK: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_EVENTBRIDGE_DETAIL_MASK: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_AWS_SES_MAIL_MASK: {
        'type': 'boolean',
        'defaultValue': True,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_AWS_SES_MAIL_DESTINATION_MASK: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_CHALICE_DISABLE: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_DJANGO_DISABLE: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_DJANGO_ORM_DISABLE: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_FLASK_DISABLE: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_FASTAPI_DISABLE: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_LOG_CONSOLE_DISABLE: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_LOG_LOGLEVEL: {
        'type': 'string',
        'defaultValue': 'TRACE',
    },
    config_names.THUNDRA_LAMBDA_DEBUGGER_ENABLE: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_LAMBDA_DEBUGGER_PORT: {
        'type': 'int',
        'defaultValue': 1111,
    },
    config_names.THUNDRA_LAMBDA_DEBUGGER_LOGS_ENABLE: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_LAMBDA_DEBUGGER_WAIT_MAX: {
        'type': 'int',
        'defaultValue': 60000,
    },
    config_names.THUNDRA_LAMBDA_DEBUGGER_IO_WAIT: {
        'type': 'int',
        'defaultValue': 2000,
    },
    config_names.THUNDRA_LAMBDA_DEBUGGER_BROKER_PORT: {
        'type': 'int',
        'defaultValue': 444,
    },
    config_names.THUNDRA_LAMBDA_DEBUGGER_BROKER_HOST: {
        'type': 'string',
        'defaultValue': 'broker.service.serverlessdebugger.com',
    },
    config_names.THUNDRA_LAMBDA_DEBUGGER_SESSION_NAME: {
        'type': 'string',
        'defaultValue': 'default',
    },
    config_names.THUNDRA_LAMBDA_DEBUGGER_AUTH_TOKEN: {
        'type': 'string',
    },
    config_names.THUNDRA_TRACE_INTEGRATIONS_TORNADO_DISABLE: {
        'type': 'boolean',
        'defaultValue': False,
    },
    config_names.THUNDRA_TEST_RUN_ID: {
        'type': 'string'
    },
    config_names.THUNDRA_TEST_PROJECT_ID: {
        'type': 'string'
    },
    config_names.THUNDRA_TEST_STATUS_REPORT_FREQUENCY: {
        'type': 'int',
        'defaultValue': 10,
    },
    config_names.THUNDRA_TEST_LOG_COUNT_MAX: {
        'type': 'int',
        'defaultValue': 100,
    },
    config_names.THUNDRA_TEST_SPAN_COUNT_MAX: {
        'type': 'int',
        'defaultValue': 100,
    },
    config_names.THUNDRA_TEST_ACTIVE: {
        'type': 'bool',
        'defaultValue': False,
    },
    config_names.THUNDRA_TEST_DISABLE: {
        'type': 'bool',
        'defaultValue': False,
    }
}



from __future__ import absolute_import

import imp
import os

from thundra.config.config_provider import ConfigProvider
from thundra.config import config_names
from thundra.thundra_agent import Thundra

thundra = Thundra()

handler_found = False
user_handler = None

handler_path = ConfigProvider.get(config_names.THUNDRA_LAMBDA_HANDLER, None)
if handler_path is None:
    raise ValueError(
        "No handler specified for \'thundra_agent_lambda_handler\' environment variable"
    )
else:
    handler_found = True
    (module_name, handler_name) = handler_path.rsplit('.', 1)
    file_handle, pathname, desc = None, None, None
    try:
        for segment in module_name.split('.'):
            if pathname is not None:
                pathname = [pathname]
            file_handle, pathname, desc = imp.find_module(segment, pathname)
        user_module = imp.load_module(module_name, file_handle, pathname, desc)
        if file_handle is None:
            module_type = desc[2]
            if module_type == imp.C_BUILTIN:
                raise ImportError("Cannot use built-in module {} as a handler module".format(module_name))
        user_handler = getattr(user_module, handler_name)
    finally:
        if file_handle:
            file_handle.close()


def wrapper(event, context):
    global user_handler
    if handler_found and user_handler:
        if not hasattr(user_handler, "thundra_wrapper"):
            user_handler = thundra(user_handler)
        return user_handler(event, context)



from threading import Lock


class WrapperFactory:
    lock = Lock()
    wrappers = {}

    @staticmethod
    def get_or_create(wrapper_class):
        with WrapperFactory.lock:
            if wrapper_class in WrapperFactory.wrappers:
                return WrapperFactory.wrappers[wrapper_class]
            else:
                WrapperFactory.wrappers[wrapper_class] = wrapper_class()
                return WrapperFactory.wrappers[wrapper_class]



import logging
import traceback
from functools import wraps

from thundra import constants
from thundra.application.global_application_info_provider import GlobalApplicationInfoProvider
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.context.execution_context_manager import ExecutionContextManager
from thundra.context.plugin_context import PluginContext
from thundra.context.tracing_execution_context_provider import TracingExecutionContextProvider
from thundra.wrappers import wrapper_utils, web_wrapper_utils
from thundra.wrappers.base_wrapper import BaseWrapper
from thundra.wrappers.flask import flask_executor

try:
    from flask import request, g
except ImportError:
    request = None
    g = None

logger = logging.getLogger(__name__)


class FlaskWrapper(BaseWrapper):

    def __init__(self, api_key=None, disable_trace=False, disable_metric=True, disable_log=True, opts=None):
        super(FlaskWrapper, self).__init__(api_key, disable_trace, disable_metric, disable_log, opts)
        self.application_info_provider = GlobalApplicationInfoProvider()
        ExecutionContextManager.set_provider(TracingExecutionContextProvider())
        self.plugin_context = PluginContext(application_info=self.application_info_provider.get_application_info(),
                                            request_count=0,
                                            executor=flask_executor,
                                            api_key=self.api_key)

        self.plugins = wrapper_utils.initialize_plugins(self.plugin_context, disable_trace, disable_metric, disable_log,
                                                        config=self.config)

        web_wrapper_utils.update_application_info(self.application_info_provider, self.plugin_context.application_info,
                                                  constants.ClassNames['FLASK'])

    def before_request(self, _request):
        # Execution context initialization
        execution_context = wrapper_utils.create_execution_context()
        execution_context.platform_data['request'] = _request

        # Execute plugin hooks before running user's handler
        self.plugin_context.request_count += 1
        self.execute_hook('before:invocation', execution_context)

        if g is not None:
            g.thundra_execution_context = execution_context

        return execution_context

    def after_request(self, response):
        try:
            if g is not None and hasattr(g, 'thundra_execution_context'):
                execution_context = g.thundra_execution_context
                if response:
                    execution_context.response = response
        except Exception as e:
            logger.error('Error setting response to context for Thundra: {}'.format(e))
        return response

    def teardown_request(self, exception=None):
        try:
            if g is not None and hasattr(g, 'thundra_execution_context'):
                execution_context = g.thundra_execution_context
                if exception:
                    execution_context.error = exception
                self.prepare_and_send_reports_async(execution_context)
        except Exception as e:
            logger.error('Error during the request teardown of Thundra: {}'.format(e))

    def __call__(self, original_func):
        if hasattr(original_func, "_thundra_wrapped") or ConfigProvider.get(config_names.THUNDRA_DISABLE, False):
            return original_func

        @wraps(original_func)
        def wrapper(*args, **kwargs):
            if request is None or getattr(request, '_thundra_wrapped', False):
                return original_func(*args, **kwargs)
            setattr(request, '_thundra_wrapped', True)
            try:
                execution_context = self.before_request(request)
            except Exception as e:
                logger.error('Error during the before part of Thundra: {}'.format(e))
                return original_func(*args, **kwargs)

            response = None
            # Invoke user handler
            try:
                response = original_func(*args, **kwargs)
                execution_context.response = response
            except Exception as e:
                try:
                    error = {
                        'type': type(e).__name__,
                        'message': str(e),
                        'traceback': traceback.format_exc()
                    }
                    self.teardown_request(error)
                except Exception as e_in:
                    logger.error("Error during the after part of Thundra: {}".format(e_in))
                raise e

            try:
                self.teardown_request()
            except Exception as e:
                logger.error("Error during the after part of Thundra: {}".format(e))
            return response

        setattr(wrapper, '_thundra_wrapped', True)
        return wrapper

    call = __call__



import time
import uuid

def current_milli_time():
    return int(time.time() * 1000)


def create_uuid4():
    return str(uuid.uuid4())


def print_debug_message_to_console(msg):
    from thundra.plugins.log.thundra_logger import debug_logger
    from thundra.config.config_provider import ConfigProvider
    from thundra.config import config_names
    if ConfigProvider.get(config_names.THUNDRA_DEBUG_ENABLE):
        debug_logger("[THUNDRA] " + msg)


from __future__ import absolute_import
from importlib import import_module

from thundra.config.config_provider import ConfigProvider
from thundra.config import config_names
import wrapt

_thundra_instance = None


def _wrapper(wrapped, _, args, kwargs):
    wrapped = _thundra_instance(wrapped)
    return wrapped(*args, **kwargs)


def patch(thundra_instance):
    if (not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_CHALICE_DISABLE)) and thundra_instance:
        global _thundra_instance
        _thundra_instance = thundra_instance
        try:
            import_module("chalice")
            wrapt.wrap_function_wrapper("chalice", "Chalice.__call__", _wrapper)
        except:
            pass



from __future__ import absolute_import

import logging

from thundra.listeners.thundra_span_listener import ThundraSpanListener

logger = logging.getLogger(__name__)

try:
    import builtins
except:
    import __builtin__ as builtins


class TagInjectorSpanListener(ThundraSpanListener):
    def __init__(self, inject_on_finish=False, tags_to_inject=None):
        self.inject_on_finish = inject_on_finish
        self.tags_to_inject = tags_to_inject

    def on_span_started(self, span):
        if not self.inject_on_finish:
            self._inject_tags(span)

    def on_span_finished(self, span):
        if self.inject_on_finish:
            self._inject_tags(span)

    def _inject_tags(self, span):
        if self.tags_to_inject:
            for tag in self.tags_to_inject:
                span.set_tag(tag, self.tags_to_inject.get(tag))

    @staticmethod
    def from_config(config):
        kwargs = {}
        inject_on_finish = config.get('injectOnFinish')
        tags_to_inject = config.get('tags')
        if inject_on_finish is not None:
            kwargs['inject_on_finish'] = inject_on_finish
        if tags_to_inject is not None:
            kwargs['tags_to_inject'] = tags_to_inject

        return TagInjectorSpanListener(**kwargs)
    
    @staticmethod
    def should_raise_exceptions():
        return False



import logging
import concurrent.futures as futures

from thundra.plugins.log.thundra_logger import debug_logger
from thundra import constants, utils
from thundra.config.config_provider import ConfigProvider
from thundra.config import config_names
from thundra.encoder import to_json

try:
    import requests
except ImportError:
    from botocore.vendored import requests

logger = logging.getLogger(__name__)


class Reporter:

    def __init__(self, api_key, session=None):
        if api_key is not None:
            self.api_key = api_key
        else:
            self.api_key = ''
            logger.error('Please set an API key!')

        if not session:
            session = requests.Session()
            adapter = requests.adapters.HTTPAdapter(pool_maxsize=20)
            session.mount("https://", adapter)

        self.session = session
        self.pool = futures.ThreadPoolExecutor(max_workers=20)

    def send_reports(self, reports, **opts):
        if not self.api_key:
            debug_logger("API key not set, not sending report to Thundra.")
            return []

        headers = {
            'Content-Type': 'application/json',
            'Authorization': 'ApiKey ' + self.api_key
        }

        base_url = self.get_collector_url()
        request_url = base_url + constants.PATH

        reports_json = self.prepare_report_json(reports)
        responses = []
        if len(reports_json) > 0:
            _futures = [self.pool.submit(self.send_batch, (request_url, headers, data)) for data in reports_json]
            responses = [future.result() for future in futures.as_completed(_futures)]

        if ConfigProvider.get(config_names.THUNDRA_DEBUG_ENABLE):
            debug_logger("Thundra API responses: " + str(responses))
        return responses

    def send_batch(self, args):
        url, headers, data = args
        return self.session.post(url, data=data, headers=headers, timeout=constants.DEFAULT_REPORT_TIMEOUT)

    def get_report_batches(self, reports):
        batch_size = ConfigProvider.get(config_names.THUNDRA_REPORT_REST_COMPOSITE_BATCH_SIZE)
        batches = [reports[i:i + batch_size] for i in range(0, len(reports), batch_size)]
        return batches


    def prepare_report_json(self, reports):
        batches = self.get_report_batches(reports)
        batched_reports = []
        for batch in batches:
            report_jsons = []
            for report in batch:
                try:
                    report_jsons.append(to_json(report, separators=(',', ':')))
                except TypeError:
                    logger.error(("Couldn't dump report with type {} to json string, "
                                  "probably it contains a byte array").format(report.get('type')))
            json_string = "[{}]".format(','.join(report_jsons))
            batched_reports.append(json_string)
        return batched_reports


    @staticmethod
    def get_collector_url():
        use_local = ConfigProvider.get(config_names.THUNDRA_REPORT_REST_LOCAL)

        if use_local:
            return 'http://' + constants.LOCAL_COLLECTOR_ENDPOINT + '/v1'
        return ConfigProvider.get(config_names.THUNDRA_REPORT_REST_BASEURL, 'https://' + utils.get_nearest_collector() + '/v1')




import uuid
import json
import logging

from thundra import constants, utils
from thundra.application.application_info_provider import ApplicationInfoProvider
from thundra.encoder import to_json
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
import thundra.wrappers.wrapper_utils as wrapper_utils

logger = logging.getLogger(__name__)


def on_finish(execution_context, request, response, span):
    try:
        if is_triggered_from_catchpoint(request):
            on_catchpoint_request_finish(execution_context, request, response, span)
    except Exception as e:
        logger.error('ERROR: {}'.format(e))


def is_triggered_from_catchpoint(request):
    if request and request.headers:
        headers = request.headers
        user_agent = headers.get(constants.HTTPHeaders['USER_AGENT'])
        return 'Catchpoint' in user_agent if user_agent else False
    return False


def on_catchpoint_request_finish(execution_context, request, response, span):
    api_key = ConfigProvider.get(config_names.THUNDRA_APIKEY, None)
    headers = request.headers if request and request.headers else {}
    region_name = headers.get(constants.CatchpointHeaders.get('REGION_NAME'))
    country_name = headers.get(constants.CatchpointHeaders.get('COUNTRY_NAME'))
    city_name = headers.get(constants.CatchpointHeaders.get('CITY_NAME'))
    test_id = headers.get(constants.CatchpointHeaders.get('TEST_ID'))
    application_name = generate_catchpoint_application_name(region_name, country_name, city_name)
    application_region = region_name if region_name else ''
    trace_id = span.context.trace_id
    transaction_id = utils.generate_id()
    span_id = utils.generate_id()
    start_timestamp = span.start_time
    finish_timestamp = span.finish_time
    duration = finish_timestamp - start_timestamp
    application_info = get_catchpoint_application_info(application_name, application_region)

    # create catchpoint invocation data
    resource = get_catchpoint_request_resource(execution_context, request, span, duration)
    invocation_data = create_catchpoint_request_invocation(execution_context, application_info, region_name,
                                                           country_name, city_name, test_id, trace_id, transaction_id,
                                                           span_id, start_timestamp, finish_timestamp, resource)
    execution_context.report(create_report_data(api_key, 'Invocation', invocation_data))

    # create catchpoint span data
    span_data = create_catchpoint_request_span(application_info, span, resource, region_name, country_name, city_name, 
                                               test_id, trace_id, transaction_id, span_id, start_timestamp, finish_timestamp)
    execution_context.report(create_report_data(api_key, 'Span', span_data))

    if response and hasattr(response, 'headers'):
        if response.headers is None:
            response.headers = {}
        response.headers[constants.THUNDRA_TRACE_ID_KEY] = span.context.trace_id


def create_catchpoint_request_invocation(execution_context, application_info, region_name, country_name, city_name,
                                         test_id, trace_id, transaction_id, span_id, start_timestamp, finish_timestamp,
                                         resource):
    if not execution_context.transaction_id:
        execution_context.transaction_id = str(uuid.uuid4())
    invocation_data = {
        'id': str(uuid.uuid4()),
        'type': 'Invocation',
        'agentVersion': constants.CatchpointProperties.get('AGENT_VERSION'),
        'dataModelVersion': constants.DATA_FORMAT_VERSION,
        'traceId': trace_id,
        'transactionId': transaction_id,
        'spanId': span_id,
        'applicationPlatform': '',
        'applicationRegion': region_name if region_name else '',
        'duration': finish_timestamp - start_timestamp,
        'startTimestamp': start_timestamp,
        'finishTimestamp': finish_timestamp,
        'coldStart': False,
        'timeout': False,
        'tags': {
            constants.CatchpointTags.get('REGION_NAME'): region_name,
            constants.CatchpointTags.get('COUNTRY_NAME'): country_name,
            constants.CatchpointTags.get('CITY_NAME'): city_name,
            constants.CatchpointTags.get('TEST_ID'): test_id
        },
        'resources': [resource],
        'userTags': {},
        'incomingTraceLinks': [],
        'outgoingTraceLinks': []
    }

    inject_application_info(invocation_data, application_info)
    if execution_context.error:
        wrapper_utils.set_error(invocation_data, execution_context.error)
    else:
        invocation_data['erroneous'] = False
        invocation_data['errorType'] = ''
        invocation_data['errorMessage'] = ''
        invocation_data['errorStack'] = ''
        invocation_data['errorCode'] = -1
    return invocation_data


def create_catchpoint_request_span(application_info, root_span, resource, region_name, country_name,
                                   city_name, test_id, trace_id, transaction_id, span_id, start_timestamp,
                                   finish_timestamp):
        span_data = {
            'id': span_id,
            'type': 'Span',
            'agentVersion': constants.CatchpointProperties.get('AGENT_VERSION'),
            'dataModelVersion': constants.DATA_FORMAT_VERSION,
            'domainName': constants.CatchpointProperties.get('HTTP_REQUEST_DOMAIN_NAME'),
            'className': constants.CatchpointProperties.get('HTTP_REQUEST_CLASS_NAME'),
            'serviceName': application_info.get('applicationName'),
            'traceId': trace_id,
            'transactionId': transaction_id,
            'spanOrder': 0,
            'operationName': resource.get('resourceName'),
            'duration': finish_timestamp - start_timestamp,
            'startTimestamp': start_timestamp,
            'finishTimestamp': finish_timestamp,
            'tags': {
                constants.HttpTags.get('HTTP_URL'): root_span.get_tag(constants.HttpTags.get('HTTP_URL')),
                constants.HttpTags.get('HTTP_HOST'): root_span.get_tag(constants.HttpTags.get('HTTP_HOST')),
                constants.HttpTags.get('HTTP_PATH'): root_span.get_tag(constants.HttpTags.get('HTTP_PATH')),
                constants.HttpTags.get('HTTP_METHOD'): root_span.get_tag(constants.HttpTags.get('HTTP_METHOD')),
                constants.HttpTags.get('QUERY_PARAMS'): root_span.get_tag(constants.HttpTags.get('QUERY_PARAMS')),
                constants.HttpTags.get('HTTP_STATUS'): root_span.get_tag(constants.HttpTags.get('HTTP_STATUS')),
                constants.CatchpointTags.get('REGION_NAME'): region_name,
                constants.CatchpointTags.get('COUNTRY_NAME'): country_name,
                constants.CatchpointTags.get('CITY_NAME'): city_name,
                constants.CatchpointTags.get('TEST_ID'): test_id
            }
        }
        inject_application_info(span_data, application_info)
        return span_data


def inject_application_info(data, application_info):
    data.update(application_info)
    data['applicationRuntime'] = ""
    data['applicationRuntimeVersion'] = ""


def get_catchpoint_application_info(application_name, application_region):
    application_id = constants.CatchpointProperties.get('APP_ID_TEMPLATE')
    application_id = application_id.replace(constants.CatchpointProperties.get('APP_NAME_PLACEHOLDER'),
                                                application_name)
    application_id = application_id.replace(constants.CatchpointProperties.get('APP_REGION_PLACEHOLDER'),
                                                application_region)
    return {
        'applicationId': application_id,
        'applicationInstanceId': utils.generate_id_from(application_id),
        'applicationName': application_name,
        'applicationClassName': constants.CatchpointProperties.get('APP_CLASS_NAME'),
        'applicationDomainName': constants.CatchpointProperties.get('APP_DOMAIN_NAME'),
        'applicationRegion': application_region,
        'applicationVersion': '',
        'applicationRuntime': None,
        'applicationRuntimeVersion': None,
        'applicationStage': '',
        'applicationTags': {},
    }


def get_catchpoint_request_resource(execution_context, request, span, duration):
    error = execution_context.error
    operation_name = execution_context.trigger_operation_name
    if operation_name is None:
        operation_name = span.operation_name
    return {
        'resourceType': constants.CatchpointProperties.get('HTTP_REQUEST_CLASS_NAME'),
        'resourceName': operation_name,
        'resourceOperation': request.method,
        'resourceCount': 1,
        'resourceErrorCount': 1 if error else 0,
        'resourceErrors': error.get('type') if error and 'type' in error else None,
        'resourceDuration': duration,
        'resourceMaxDuration': duration,
        'resourceAvgDuration': duration
    }


def generate_catchpoint_application_name(region_name, country_name, city_name):
    if city_name:
        return city_name
    elif country_name:
        return country_name
    elif region_name:
        return region_name
    else:
        return constants.CatchpointProperties.get('DEFAULT_APP_NAME')


def create_report_data(api_key, data_type, data):
    report_data = {
        'apiKey': api_key,
        'type': data_type,
        'dataModelVersion': constants.DATA_FORMAT_VERSION,
        'data': data
    }
    return json.loads(to_json(report_data))



from thundra.config.config_provider import ConfigProvider
from thundra.config import config_names
import foresight.utils.generic_utils as utils
import uuid, logging

LOGGER = logging.getLogger(__name__) 

class NULL_NAMESPACE:
    bytes = b''

class TestRunnerUtils:
    
    @staticmethod
    def get_configured_test_run_id():
        return ConfigProvider.get(config_names.THUNDRA_TEST_RUN_ID)

    
    @classmethod
    def get_test_run_id(cls, environment, repo_url, commit_hash, test_run_key):
        try:
            test_run_id_seed = cls.string_concat_by_underscore(environment, repo_url, 
                commit_hash, test_run_key)
            # UUID.nameUUIDFromBytes is used in java and its uuid version is 3. To keep compatibality, using uuid3.
            # https://stackoverflow.com/questions/27939281/reproduce-uuid-from-java-code-in-python
            return str(uuid.uuid3(NULL_NAMESPACE, test_run_id_seed))
        except Exception as err:
            LOGGER.error("Test run id could not be created by uuid: {}".format(err))
            pass
        return str(uuid.uuid4())


    @staticmethod
    def get_default_test_run_id(environment=None, repo_url=None, commit_hash=None):
        try:
            #TODO Find a way to generate uuid with parameters
            return utils.create_uuid4()
        except Exception as err:
            LOGGER.error("get_default_test_run_id couldn't created")
            pass


    @staticmethod
    def string_concat_by_underscore(*str_list):
        try:
            return "_".join(str_list)
        except Exception as err:
            LOGGER.error("test runner utils string concat error: {}".format(err))
            pass
        return ""


class PluginContext:

    def __init__(self, **opts):
        self.application_info = opts.get('application_info')
        self.request_count = opts.get('request_count', 0)
        self.api_key = opts.get('api_key')
        self.executor = opts.get('executor')



import sys

_ver = sys.version_info

# Python 2.x
PY2 = (_ver[0] == 2)

# Python 3.x
PY3 = (_ver[0] == 3)

# Python 2.7.x
PY27 = (PY2 and _ver[1] == 7)

# Python 3.6.x
PY36 = (PY3 and _ver[1] == 6)

# Python 3.7.x
PY37 = (PY3 and _ver[1] == 7)

# Python 3.8.x
PY38 = (PY3 and _ver[1] == 8)

class TimeoutError(Exception):
    def __init__(self, msg="Task timed out"):
        super(TimeoutError, self).__init__(msg)

if PY2:
    from urlparse import urlparse

    builtin_str = str
    bytes = str
    str = unicode
    basestring = basestring

    import Queue
    queue = Queue


elif PY3:
    from urllib.parse import urlparse

    builtin_str = str
    str = str
    bytes = bytes
    basestring = (str, bytes)
    numeric_types = (int, float)

    import queue



from threading import Lock

from thundra import constants
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.samplers.base_sampler import BaseSampler


class CountAwareSampler(BaseSampler):

    def __init__(self, count_freq=None):
        freq_from_env = ConfigProvider.get(config_names.THUNDRA_SAMPLER_COUNTAWARE_COUNTFREQ, -1)
        if freq_from_env > 0:
            self.count_freq = freq_from_env
        elif count_freq is not None:
            self.count_freq = count_freq
        else:
            self.count_freq = constants.DEFAULT_METRIC_SAMPLING_COUNT_FREQ

        self._counter = -1
        self._lock = Lock()

    def is_sampled(self, args=None):
        return self._increment_and_get_counter() % self.count_freq == 0

    def _increment_and_get_counter(self):
        with self._lock:
            self._counter += 1

        return self._counter



from foresight.pytest_integration.pytest_helper import PytestHelper
from foresight.test_status import increase_actions, TestStatus
import foresight.pytest_integration.constants as constants
import wrapt, logging, traceback


logger = logging.getLogger(__name__)

class XPassedException(Exception):
    pass


def check_test_case_result(item, execution_context, result, exception):
    """ Handle test case result.

    Args:
        item (pytest.Item): Current running test case.
        execution_context (TestCaseExecutionContext): Thundra execution context
        result (pytest.Result): test case result
        exception (pytest.ExceptionInfo): ExceptionInfo for test case if any. 

    Returns:
        str: One of the result in TestStatus class.
    """
    test_status = TestStatus.SKIPPED
    try:
        xfail = (hasattr(result, "wasxfail") or "xfail" in result.keywords) and check_and_get_xfail_condition(item)
        has_skip_keyword = any(x in result.keywords for x in ["skip", "skipif", "skipped"])

        if hasattr(item, constants.THUNDRA_MARKED_AS_SKIPPED):
            delattr(item, constants.THUNDRA_MARKED_AS_SKIPPED)
            test_status = TestStatus.SKIPPED
        elif (exception and hasattr(exception.value, "msg") and 
            "timeout" in exception.value.msg.lower()):
            test_status = TestStatus.ABORTED
            handle_error(exception, result, execution_context)
            execution_context.timeout = True
        elif result.skipped:
            if xfail and not has_skip_keyword:
                test_status = TestStatus.SUCCESSFUL
            else:
                test_status = TestStatus.SKIPPED
        elif result.passed:
            if xfail and not has_skip_keyword:
                if exception == None:
                    exception = XPassedException(f"{result.nodeid} test result has been passed while expecting to get xfail!")
                test_status = TestStatus.FAILED
                handle_xpass(exception, result, execution_context)
            else:
                test_status = TestStatus.SUCCESSFUL
        else: # failed
            test_status = TestStatus.FAILED
            handle_error(exception, result, execution_context)
    except Exception as err:
        logger.error("Couldn't set test status properly for pytest: {}".format(err))
        pass
    return test_status

def check_and_get_xfail_condition(item):
    """Check test item that marked as xfail and get its xfail condition if it exists.

    Args:
        item (pytest.Item): Test case item that contains all info about current test.

    Returns:
        bool: If condition is string, then first evaluate and return result. If condition is bool, return
            immediately the bool. If there is an exception, always returns True.
    """
    try:
        own_markers = item.own_markers
        for i in own_markers:
            if i.name == "xfail" and len(i.args) > 0:
                condition_xfail = i.args[0]
                if isinstance(condition_xfail, str):
                    from argparse import Namespace
                    import os, sys
                    result = eval(condition_xfail, Namespace(globals=globals(), os=os, sys=sys, config=item.config).__dict__)
                    return result
                elif isinstance(condition_xfail, bool):
                    if condition_xfail == False:
                        return False
                    else:
                        return True
                else:
                    break
    except Exception as e:
        logger.error("Check and get xfail condition error: {}".format(e))
        pass
    return True


def update_test_status(item, test_status, execution_context):
    """update successful, skipped, aborted and total test status  w.r.t test case result calculated in 
    check_test_case_result function.

    Args:
        item (pytest.Item): Current running test case
        test_status (str): Current test case result
        execution_context (TestCaseExecutionContext): Execution context for test case
    """
    try:
        increase_action = increase_actions[test_status]
        execution_context.set_status(test_status)
        if increase_action:
            setattr(item, constants.THUNDRA_TEST_RESULTED, True)
            increase_action()
    except Exception as err:
        logger.error("Couldn't update test status for pytest: {}".format(err))
        pass


def handle_xpass(exception, result, execution_context):
    """Set exception into execution context if any.

    Args:
        exception (pytest.ExceptionInfo): Test case Exception info
        result (pytest.Result): Test result 
        execution_context (TestCaseExecutionContext): Execution context for test case
    """
    try:
        execution_context.error = {
                        'type': type(exception).__name__,
                        'message': result.longreprtext or str(exception), # TODO longreprtext can be empty. Search what can be used for it.
                        'traceback': ''
                    }
    except Exception as err:
        logger.error("Couldn't handle error for pytest: {}".format(err))
        pass

def handle_error(exception, result, execution_context):
    """Set exception into execution context if any.

    Args:
        exception (pytest.ExceptionInfo): Test case Exception info
        result (pytest.Result): Test result 
        execution_context (TestCaseExecutionContext): Execution context for test case
    """
    try:
        execution_context.error = {
                        'type': type(exception).__name__,
                        'message': result.longreprtext or str(exception), # TODO longreprtext can be empty. Search what can be used for it.
                        'traceback': ''.join(traceback.format_tb(exception.tb))
                    }
    except Exception as err:
        logger.error("Couldn't handle error for pytest: {}".format(err))
        pass


def set_attributes_test_item(item, module_item):
    """Set current test case item for thundra trace. 

    Args:
        item (pytest.Item): Current running test case.
    """
    try:
        setattr(item, "scope", "function")  
        setattr(module_item, "scope", "module")
        own_markers = item.own_markers
        check_marked_as_skipped = any("skip" in mark.name for mark in own_markers)
        if check_marked_as_skipped:
            setattr(item, constants.THUNDRA_MARKED_AS_SKIPPED, True)
    except Exception as err:
        logger.error("Couldn't set attributets test item for pytest: {}".format(err))
        pass


def check_test_status_state(item, call):
    """Check current running test case state and return status and exception if any.

    Args:
        item (pytest.Item): Current running test case
        call (pytest.CallInfo): Result/Exception info a function invocation.

    Returns:
        bool: True if test status should be calculated o.w False.
        pytest.ExceptionInfo: Exception info if any else None
    """
    try:
        is_setup_or_teardown = call.when == 'setup' or call.when == 'teardown'
        exception = call.excinfo
        status = True
        if (is_setup_or_teardown and not exception) or hasattr(item, constants.THUNDRA_TEST_RESULTED):
            status = False
        '''
            Pytest not run call when function skipped. So, if function has THUNDRA_MARKED_AS_SKIPPED and
            call.when is call state, then condition for skipif returns false and just evalute result with default.
        '''
        if call.when == "call" and hasattr(item, constants.THUNDRA_MARKED_AS_SKIPPED):
            delattr(item, constants.THUNDRA_MARKED_AS_SKIPPED)
        return status, exception
    except Exception as err:
        logger.error("Couldn't check test status state: {}".format(err))
        pass
    return False, None


def _check_thundra_fixture(request):
    return constants.THUNDRA_FIXTURE_PREFIX in request.fixturename


def _check_request_scope_function(request):
    return request.scope == "function"


def _check_request_scope_fixture(request): #TODO for now, class fixture do not support!!!
    return request.scope == "function" or request.scope == "module"


def fixture_closure(request, setup_or_teardown, start_or_finish):
    """Handle beforeeach, aftereach, beforeall and afterall process for thundra.

    Args:
        request (pytest.FixtureRequest): The request fixture is a special fixture providing information of the requesting test function.
        setup_or_teardown (bool, optional): To decide current state is fixture setup or teardown.
        start_or_finish (bool, optional): To decide span should be started or finished. 
    """
    def handle_fixture_setup():
        if start_or_finish:
            if _check_request_scope_function(request):
                PytestHelper.start_test_span(request.node)
                PytestHelper.start_before_each_span(request)
            else:
                PytestHelper.start_before_all_span(request)
        else:
            if _check_request_scope_function(request):
                PytestHelper.finish_before_each_span(request)
            else:
                PytestHelper.finish_before_all_span(request)  


    def handle_fixture_teardown():
        if start_or_finish:
            if _check_request_scope_function(request):
                PytestHelper.start_after_each_span(request)
            else:
                PytestHelper.start_after_all_span(request)
        else:
            if _check_request_scope_function(request):
                PytestHelper.finish_after_each_span(request)
            else:
                PytestHelper.finish_after_all_span(request)


    if _check_thundra_fixture(request) or not request:
        return
    else:
        if setup_or_teardown:
            handle_fixture_setup()
        else:
            handle_fixture_teardown()



def _wrapper_setup_fixture(wrapped, instance, args, kwargs):
    res = None
    request = None
    try:
        request = args[1]
    except Exception as err:
        logger.error("Couldn't get request in wrapper_setup function: {}".format(err))
    if _check_request_scope_fixture(request):
        fixture_closure(request, setup_or_teardown=True, start_or_finish=True)
        res = wrapped(*args, **kwargs)
        fixture_closure(request, setup_or_teardown=True, start_or_finish=False)
    else:
        res = wrapped(*args, **kwargs)
    if res != None:
        return res



def _wrapper_teardown_fixture(wrapped, instance, args, kwargs):
    request = None
    try:
        request = kwargs["request"]
    except Exception as err:
        logger.error("Couldn't get request in wrapper_teardown function: {}".format(err))
    '''
        Pytest stores fixtures teardown functions in _finalizers list. If there is no 
        teardown function no need to create span.
    '''
    if instance._finalizers and _check_request_scope_fixture(request):
        fixture_closure(request, setup_or_teardown=False, start_or_finish=True)
        wrapped(*args, **kwargs)
        fixture_closure(request, setup_or_teardown=False, start_or_finish=False)
    else:
        wrapped(*args, **kwargs)

def patch():
    '''
        fixture function has been called in call_fixture_func.
    '''
    wrapt.wrap_function_wrapper(
            "_pytest.fixtures",
            "call_fixture_func",
            _wrapper_setup_fixture
        )
    '''
        teardown functions has been stored in a stack(FixtureDef._finalizer).
        finish function has been iterated over this stack and call teardown fixtures.
    '''    
    wrapt.wrap_function_wrapper(
            "_pytest.fixtures",
            "FixtureDef.finish",
            _wrapper_teardown_fixture
        )


class ApplicationInfo:

    def __init__(self, application_id = None, application_instance_id = None, application_domain_name = None,
         application_class_name = None, application_name = None, application_version = None, 
         application_stage = None, application_runtime = None, 
         application_runtime_version = None, application_tags = None):
        
        self.application_id = application_id 
        self.application_instance_id = application_instance_id 
        self.application_domain_name = application_domain_name 
        self.application_class_name = application_class_name 
        self.application_name = application_name 
        self.application_version = application_version 
        self.application_stage = application_stage 
        self.application_runtime = application_runtime 
        self.application_runtime_version = application_runtime_version 
        self.applicationTags = application_tags

    def to_json(self):
        return {
            'applicationId': self.application_id,
            'applicationInstanceId': self.application_instance_id,
            'applicationDomainName': self.application_domain_name,
            'applicationClassName': self.application_class_name,
            'applicationName': self.application_name,
            'applicationVersion': self.application_version,
            'applicationStage': self.application_stage,
            'applicationRuntime': self.application_runtime,
            'applicationRuntimeVersion': self.application_runtime_version,
            'applicationTags': self.applicationTags
        }


import codecs
import os.path

from setuptools import setup, find_packages


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


setup(name='thundra',
      version=get_version('thundra/_version.py'),
      description='Thundra Python agent',
      long_description='Thundra Python agent',
      url='https://github.com/thundra-agent-python',
      author='Thundra',
      author_email='python@thundra.io',
      python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*',
      packages=find_packages(exclude=('tests', 'tests.*',)),
      install_requires=['requests>=2.16.0', 'opentracing>=2.0', 'wrapt>=1.10.11', 'simplejson', 'enum-compat',
                        'jsonpickle==1.3', 'websocket-client', 'python-dateutil', 'GitPython>=3.1.18', 'fastcounter>=1.1.0', 'pympler'],
      zip_safe=True,
      entry_points={"pytest11": ["thundra-foresight = foresight.pytest_integration.plugin"]},
      classifiers=[
          "Development Status :: 5 - Production/Stable",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: Apache Software License",
          "Programming Language :: Python",
          "Programming Language :: Python :: 2",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Programming Language :: Python :: 3.8",
      ],
      )
